#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Composite classes for the VIIRS instrument."""

from __future__ import annotations

import datetime as dt
import logging
import math

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import combine_metadata

from .core import CompositeBase, GenericCompositor

LOG = logging.getLogger(__name__)


class HistogramDNB(CompositeBase):
    """Histogram equalized DNB composite.

    The logic for this code was taken from Polar2Grid and was originally developed by Eva Schiffer (SSEC).

    This composite separates the DNB data in to 3 main regions: Day, Night, and Mixed. Each region is
    equalized separately to bring out the most information from the region due to the high dynamic range
    of the DNB data. Optionally, the mixed region can be separated in to multiple smaller regions by
    using the `mixed_degree_step` keyword.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the compositor with values from the user or from the configuration file.

        :param high_angle_cutoff: solar zenith angle threshold in degrees, values above this are considered "night"
        :param low_angle_cutoff: solar zenith angle threshold in degrees, values below this are considered "day"
        :param mixed_degree_step: Step interval to separate "mixed" region in to multiple parts
                                  by default does whole mixed region
        """
        self.high_angle_cutoff = int(kwargs.pop("high_angle_cutoff", 100))
        self.low_angle_cutoff = int(kwargs.pop("low_angle_cutoff", 88))
        self.mixed_degree_step = int(kwargs.pop(
            "mixed_degree_step")) if "mixed_degree_step" in kwargs else None
        super(HistogramDNB, self).__init__(*args, **kwargs)

    def __call__(self, datasets, **info):
        """Create the composite by scaling the DNB data using a histogram equalization method.

        Args:
            datasets: 2-element tuple (Day/Night Band data, Solar Zenith Angle data)
            info: Miscellaneous metadata for the newly produced composite
        """
        if len(datasets) != 2:
            raise ValueError("Expected 2 datasets, got %d" % (len(datasets), ))

        dnb_data = datasets[0]
        sza_data = datasets[1]
        output_data = da.map_blocks(
            self._run_dnb_normalization,
            dnb_data.data.rechunk(dnb_data.shape),
            sza_data.data.rechunk(sza_data.shape),
            dtype=dnb_data.dtype,
            meta=np.ndarray((), dtype=dnb_data.dtype),
        )
        output_dataset = dnb_data.copy()
        output_dataset.data = output_data.rechunk(dnb_data.data.chunks)

        info = dnb_data.attrs.copy()
        info.update(self.attrs)
        info["standard_name"] = "equalized_radiance"
        info["mode"] = "L"
        output_dataset.attrs = info
        return output_dataset

    def _run_dnb_normalization(self, dnb_data: np.ndarray, sza_data: np.ndarray) -> np.ndarray:
        """Scale the DNB data using a histogram equalization method.

        Args:
            dnb_data: Day/Night Band data numpy array
            sza_data: Solar Zenith Angle data numpy array

        """
        # convert dask arrays to DataArray objects
        dnb_data_arr = xr.DataArray(dnb_data, dims=("y", "x"))
        sza_data_arr = xr.DataArray(sza_data, dims=("y", "x"))

        good_mask = ~(dnb_data_arr.isnull() | sza_data_arr.isnull())
        output_data_arr = dnb_data_arr.where(good_mask)
        # we only need the numpy array
        output_data = output_data_arr.values.copy()
        dnb_data = dnb_data_arr.values
        sza_data = sza_data_arr.values
        self._normalize_dnb_for_mask(dnb_data, sza_data, good_mask, output_data)
        return output_data

    def _normalize_dnb_for_mask(self, dnb_data, sza_data, good_mask, output_dataset):
        day_mask, mixed_mask, night_mask = make_day_night_masks(
            sza_data,
            good_mask.values,
            self.high_angle_cutoff,
            self.low_angle_cutoff,
            stepsDegrees=self.mixed_degree_step)
        self._normalize_dnb_with_day_night_masks(dnb_data, day_mask, mixed_mask, night_mask, output_dataset)

    def _normalize_dnb_with_day_night_masks(self, dnb_data, day_mask, mixed_mask, night_mask, output_dataset):
        histogram_equalization(dnb_data, day_mask, out=output_dataset)
        for mask in mixed_mask:
            histogram_equalization(dnb_data, mask, out=output_dataset)
        histogram_equalization(dnb_data, night_mask, out=output_dataset)


class AdaptiveDNB(HistogramDNB):
    """Adaptive histogram equalized DNB composite.

    The logic for this code was taken from Polar2Grid and was originally developed by Eva Schiffer (SSEC).

    This composite separates the DNB data in to 3 main regions: Day, Night, and Mixed. Each region is
    equalized separately to bring out the most information from the region due to the high dynamic range
    of the DNB data. Optionally, the mixed region can be separated in to multiple smaller regions by
    using the `mixed_degree_step` keyword.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the compositor with values from the user or from the configuration file.

        Adaptive histogram equalization and regular histogram equalization can be configured
        independently for each region: day, night, or mixed.
        A region can be set to use adaptive equalization "always", or "never", or only when
        there are multiple regions in a single scene "multiple" via the `adaptive_X` keyword
        arguments (see below).

        :param adaptive_day: one of ("always", "multiple", "never") meaning when adaptive equalization is used.
        :param adaptive_mixed: one of ("always", "multiple", "never") meaning when adaptive equalization is used.
        :param adaptive_night: one of ("always", "multiple", "never") meaning when adaptive equalization is used.
        """
        self.adaptive_day = kwargs.pop("adaptive_day", "always")
        self.adaptive_mixed = kwargs.pop("adaptive_mixed", "always")
        self.adaptive_night = kwargs.pop("adaptive_night", "always")
        self.day_radius_pixels = int(kwargs.pop("day_radius_pixels", 400))
        self.mixed_radius_pixels = int(kwargs.pop("mixed_radius_pixels", 100))
        self.night_radius_pixels = int(kwargs.pop("night_radius_pixels", 400))

        super(AdaptiveDNB, self).__init__(*args, **kwargs)

    def _normalize_dnb_for_mask(self, dnb_data, sza_data, good_mask, output_dataset):
        day_mask, mixed_mask, night_mask = make_day_night_masks(
            sza_data,
            good_mask.values,
            self.high_angle_cutoff,
            self.low_angle_cutoff,
            stepsDegrees=self.mixed_degree_step)

        has_multi_times = len(mixed_mask) > 0
        if self.adaptive_day == "always" or (
                has_multi_times and self.adaptive_day == "multiple"):
            LOG.debug("Adaptive histogram equalizing DNB day data...")
            local_histogram_equalization(
                dnb_data,
                day_mask,
                valid_data_mask=good_mask.values,
                local_radius_px=self.day_radius_pixels,
                out=output_dataset)
        else:
            LOG.debug("Histogram equalizing DNB day data...")
            histogram_equalization(dnb_data,
                                   day_mask,
                                   out=output_dataset)
        for mask in mixed_mask:
            if self.adaptive_mixed == "always" or (
                    has_multi_times and
                    self.adaptive_mixed == "multiple"):
                LOG.debug(
                    "Adaptive histogram equalizing DNB mixed data...")
                local_histogram_equalization(
                    dnb_data,
                    mask,
                    valid_data_mask=good_mask.values,
                    local_radius_px=self.mixed_radius_pixels,
                    out=output_dataset)
            else:
                LOG.debug("Histogram equalizing DNB mixed data...")
                histogram_equalization(dnb_data,
                                       day_mask,
                                       out=output_dataset)
        if self.adaptive_night == "always" or (
                has_multi_times and self.adaptive_night == "multiple"):
            LOG.debug("Adaptive histogram equalizing DNB night data...")
            local_histogram_equalization(
                dnb_data,
                night_mask,
                valid_data_mask=good_mask.values,
                local_radius_px=self.night_radius_pixels,
                out=output_dataset)
        else:
            LOG.debug("Histogram equalizing DNB night data...")
            histogram_equalization(dnb_data,
                                   night_mask,
                                   out=output_dataset)


class ERFDNB(CompositeBase):
    """Equalized DNB composite using the error function (erf).

    The logic for this code was taken from Polar2Grid and was originally
    developed by Curtis Seaman and Steve Miller. The original code was
    written in IDL and is included as comments in the code below.

    """

    def __init__(self, *args, **kwargs):
        """Initialize ERFDNB specific keyword arguments."""
        self.saturation_correction = kwargs.pop("saturation_correction",
                                                False)
        super(ERFDNB, self).__init__(*args, **kwargs)

    def _saturation_correction(self, dnb_data, min_val, max_val, unit_factor):
        saturation_pct = float(np.count_nonzero(dnb_data >
                                                max_val)) / dnb_data.size
        LOG.debug("Dynamic DNB saturation percentage: %f", saturation_pct)
        while saturation_pct > 0.005:
            max_val *= 1.1
            saturation_pct = float(np.count_nonzero(
                dnb_data > max_val)) / dnb_data.size
            LOG.debug("Dynamic DNB saturation percentage: %f",
                      saturation_pct)

        inner_sqrt = (dnb_data - min_val) / (max_val - min_val)
        # clip negative values to 0 before the sqrt
        inner_sqrt[inner_sqrt < 0] = 0
        return np.sqrt(inner_sqrt)

    def __call__(self, datasets, **info):
        """Create the composite DataArray object for ERFDNB."""
        if len(datasets) != 4:
            raise ValueError("Expected 4 datasets, got %d" % (len(datasets), ))

        from scipy.special import erf
        dnb_data = datasets[0]
        sza_data = datasets[1]
        lza_data = datasets[2]
        output_dataset = dnb_data.where(~(dnb_data.isnull() | sza_data.isnull()))
        # this algorithm assumes units of "W cm-2 sr-1" so if there are other
        # units we need to adjust for that
        if dnb_data.attrs.get("units", "W m-2 sr-1") == "W m-2 sr-1":
            unit_factor = 10000.
        else:
            unit_factor = 1.

        # convert to decimal instead of %
        moon_illum_fraction = da.mean(datasets[3].data) * 0.01

        # From Steve Miller and Curtis Seaman
        # maxval = 10.^(-1.7 - (((2.65+moon_factor1+moon_factor2))*(1+erf((solar_zenith-95.)/(5.*sqrt(2.0))))))
        # minval = 10.^(-4. - ((2.95+moon_factor2)*(1+erf((solar_zenith-95.)/(5.*sqrt(2.0))))))
        # scaled_radiance = (radiance - minval) / (maxval - minval)
        # radiance = sqrt(scaled_radiance)

        # Version 2: Update from Curtis Seaman
        # maxval = 10.^(-1.7 - (((2.65+moon_factor1+moon_factor2))*(1+erf((solar_zenith-95.)/(5.*sqrt(2.0))))))
        # minval = 10.^(-4. - ((2.95+moon_factor2)*(1+erf((solar_zenith-95.)/(5.*sqrt(2.0))))))
        # saturated_pixels = where(radiance gt maxval, nsatpx)
        # saturation_pct = float(nsatpx)/float(n_elements(radiance))
        # print, 'Saturation (%) = ', saturation_pct
        #
        # while saturation_pct gt 0.005 do begin
        #   maxval = maxval*1.1
        #   saturated_pixels = where(radiance gt maxval, nsatpx)
        #   saturation_pct = float(nsatpx)/float(n_elements(radiance))
        #   print, saturation_pct
        # endwhile
        #
        # scaled_radiance = (radiance - minval) / (maxval - minval)
        # radiance = sqrt(scaled_radiance)

        moon_factor1 = 0.7 * (1.0 - moon_illum_fraction)
        moon_factor2 = 0.0022 * lza_data.data
        erf_portion = 1 + erf((sza_data.data - 95.0) / (5.0 * np.sqrt(2.0)))
        max_val = da.power(
            10, -1.7 -
            (2.65 + moon_factor1 + moon_factor2) * erf_portion) * unit_factor
        min_val = da.power(10, -4.0 -
                           (2.95 + moon_factor2) * erf_portion) * unit_factor

        # Update from Curtis Seaman, increase max radiance curve until less
        # than 0.5% is saturated
        if self.saturation_correction:
            output_data = da.map_blocks(
                self._saturation_correction,
                output_dataset.data.rechunk(output_dataset.shape),
                min_val.rechunk(output_dataset.shape),
                max_val.rechunk(output_dataset.shape),
                unit_factor,
                dtype=output_dataset.dtype,
                meta=np.ndarray((), dtype=output_dataset.dtype),
            )
            output_dataset.data = output_data.rechunk(dnb_data.data.chunks)
        else:
            inner_sqrt = (output_dataset - min_val) / (max_val - min_val)
            # clip negative values to 0 before the sqrt
            inner_sqrt.data = np.clip(inner_sqrt.data, 0, None)
            output_dataset.data = np.sqrt(inner_sqrt).data

        info = dnb_data.attrs.copy()
        info.update(self.attrs)
        info["standard_name"] = "equalized_radiance"
        info["mode"] = "L"
        output_dataset.attrs = info
        return output_dataset


def make_day_night_masks(solarZenithAngle,
                         good_mask,
                         highAngleCutoff,
                         lowAngleCutoff,
                         stepsDegrees=None):
    """Generate masks for day, night, and twilight regions.

    Masks are created from the provided solar zenith angle data.

    Optionally provide the highAngleCutoff and lowAngleCutoff that define
    the limits of the terminator region (if no cutoffs are given the
    DEFAULT_HIGH_ANGLE and DEFAULT_LOW_ANGLE will be used).

    Optionally provide the stepsDegrees that define how many degrees each
    "mixed" mask in the terminator region should be (if no stepsDegrees is
    given, the whole terminator region will be one mask).
    """
    # if the caller passes None, we're only doing one step
    stepsDegrees = highAngleCutoff - lowAngleCutoff if stepsDegrees is None else stepsDegrees

    night_mask = (solarZenithAngle > highAngleCutoff) & good_mask
    day_mask = (solarZenithAngle <= lowAngleCutoff) & good_mask
    mixed_mask = []
    steps = list(range(lowAngleCutoff, highAngleCutoff + 1, stepsDegrees))
    if steps[-1] >= highAngleCutoff:
        steps[-1] = highAngleCutoff
    steps = zip(steps, steps[1:])
    for i, j in steps:
        LOG.debug("Processing step %d to %d" % (i, j))
        tmp = (solarZenithAngle > i) & (solarZenithAngle <= j) & good_mask
        if tmp.any():
            LOG.debug("Adding step %d to %d" % (i, j))
            # log.debug("Points to process in this range: " + str(np.sum(tmp)))
            mixed_mask.append(tmp)
        del tmp

    return day_mask, mixed_mask, night_mask


def histogram_equalization(
        data,
        mask_to_equalize,
        number_of_bins=1000,
        std_mult_cutoff=4.0,
        do_zerotoone_normalization=True,
        out=None):
    """Perform a histogram equalization on the data.

    Data is selected by the mask_to_equalize mask.
    The data will be separated into number_of_bins levels for equalization and
    outliers beyond +/- std_mult_cutoff*std will be ignored.

    If do_zerotoone_normalization is True the data selected by mask_to_equalize
    will be returned in the 0 to 1 range. Otherwise the data selected by
    mask_to_equalize will be returned in the 0 to number_of_bins range.

    Note: the data will be changed in place.
    """
    out = out if out is not None else data.copy()

    LOG.debug("determining DNB data range for histogram equalization")
    sub_arr = data[mask_to_equalize]
    if sub_arr.size == 0:
        # no good data
        return out

    avg = np.mean(sub_arr)
    std = np.std(sub_arr)
    # limit our range to +/- std_mult_cutoff*std; e.g. the default
    # std_mult_cutoff is 4.0 so about 99.8% of the data
    concervative_mask = (data < (avg + std * std_mult_cutoff)) & (
        data > (avg - std * std_mult_cutoff)) & mask_to_equalize

    LOG.debug("running histogram equalization")
    cumulative_dist_function, temp_bins = _histogram_equalization_helper(
        data[concervative_mask],
        number_of_bins)

    # linearly interpolate using the distribution function to get the new
    # values
    out[mask_to_equalize] = np.interp(data[mask_to_equalize], temp_bins[:-1],
                                      cumulative_dist_function)

    # if we were asked to, normalize our data to be between zero and one,
    # rather than zero and number_of_bins
    if do_zerotoone_normalization:
        _linear_normalization_from_0to1(out, mask_to_equalize, number_of_bins)

    return out


def local_histogram_equalization(data, mask_to_equalize, valid_data_mask=None, number_of_bins=1000,
                                 std_mult_cutoff=3.0,
                                 do_zerotoone_normalization=True,
                                 local_radius_px: int = 300,
                                 clip_limit=60.0,
                                 slope_limit=3.0,
                                 do_log_scale=True,
                                 # can't take the log of zero, so the offset
                                 # may be needed; pass 0.0 if your data doesn't
                                 # need it
                                 log_offset=0.00001,
                                 out=None
                                 ):
    """Equalize the provided data (in the mask_to_equalize) using adaptive histogram equalization.

    Tiles of width/height (2 * local_radius_px + 1) will be calculated and results for each pixel will be bilinearly
    interpolated from the nearest 4 tiles when pixels fall near the edge of the image (there is no adjacent tile) the
    resultant interpolated sum from the available tiles will be multiplied to account for the weight of any missing
    tiles::

        pixel total interpolated value = pixel available interpolated value / (1 - missing interpolation weight)

    If ``do_zerotoone_normalization`` is True the data will be scaled so that all data in the mask_to_equalize falls
    between 0 and 1; otherwise the data in mask_to_equalize will all fall between 0 and number_of_bins.

    Returns: The equalized data
    """
    out = out if out is not None else np.zeros_like(data)
    # if we don't have a valid mask, use the mask of what we should be equalizing
    if valid_data_mask is None:
        valid_data_mask = mask_to_equalize

    # calculate some useful numbers for our tile math
    total_rows = data.shape[0]
    total_cols = data.shape[1]
    tile_size = int(local_radius_px * 2 + 1)
    row_tiles = math.ceil(total_rows / tile_size)
    col_tiles = math.ceil(total_cols / tile_size)

    all_cumulative_dist_functions, all_bin_information = _compute_tile_dist_and_bin_info(
        data,
        valid_data_mask,
        std_mult_cutoff,
        do_log_scale,
        log_offset,
        clip_limit,
        slope_limit,
        number_of_bins,
        row_tiles,
        col_tiles,
        tile_size,
    )

    # get the tile weight array so we can use it to interpolate our data
    tile_weights = _calculate_weights(tile_size)

    # now loop through our tiles and linearly interpolate the equalized
    # versions of the data
    for num_row_tile in range(row_tiles):
        for num_col_tile in range(col_tiles):
            _interpolate_local_equalized_tiles(
                data, out, mask_to_equalize, valid_data_mask, do_log_scale, log_offset,
                tile_weights, all_bin_information, all_cumulative_dist_functions,
                num_row_tile, num_col_tile, tile_size,
            )

    # if we were asked to, normalize our data to be between zero and one,
    # rather than zero and number_of_bins
    if do_zerotoone_normalization:
        _linear_normalization_from_0to1(out, mask_to_equalize, number_of_bins)

    return out


def _compute_tile_dist_and_bin_info(
        data: np.ndarray,
        valid_data_mask: np.ndarray,
        std_mult_cutoff: float,
        do_log_scale: bool,
        log_offset: float,
        clip_limit: float,
        slope_limit: float,
        number_of_bins: int,
        row_tiles: int,
        col_tiles: int,
        tile_size: int,
):
    # an array of our distribution functions for equalization
    all_cumulative_dist_functions = []
    # an array of our bin information for equalization
    all_bin_information = []

    # loop through our tiles and create the histogram equalizations for each one
    for num_row_tile in range(row_tiles):
        row_dist_functions = []
        row_bin_info = []
        for num_col_tile in range(col_tiles):
            tile_dist_func, tile_bin_info = _histogram_equalize_one_tile(
                data, valid_data_mask, std_mult_cutoff, do_log_scale, log_offset,
                clip_limit, slope_limit, number_of_bins, num_row_tile, num_col_tile,
                tile_size
            )
            row_dist_functions.append(tile_dist_func)
            row_bin_info.append(tile_bin_info)
        all_cumulative_dist_functions.append(row_dist_functions)
        all_bin_information.append(row_bin_info)

    return all_cumulative_dist_functions, all_bin_information


def _histogram_equalize_one_tile(
        data, valid_data_mask, std_mult_cutoff, do_log_scale, log_offset,
        clip_limit, slope_limit, number_of_bins, num_row_tile, num_col_tile,
        tile_size):
    # calculate the range for this tile (min is inclusive, max is
    # exclusive)
    min_row = num_row_tile * tile_size
    max_row = min_row + tile_size
    min_col = num_col_tile * tile_size
    max_col = min_col + tile_size

    # for speed of calculation, pull out the mask of pixels that should
    # be used to calculate the histogram
    mask_valid_data_in_tile = valid_data_mask[min_row:max_row, min_col:max_col]

    # if we have any valid data in this tile, calculate a histogram equalization for this tile
    # (note: even if this tile does not fall in the mask_to_equalize, it's histogram may be used by other tiles)
    if not mask_valid_data_in_tile.any():
        return None, None

    # use all valid data in the tile, so separate sections will
    # blend cleanly
    temp_valid_data = data[min_row:max_row, min_col:max_col][
        mask_valid_data_in_tile]
    temp_valid_data = temp_valid_data[
        temp_valid_data >= 0
        ]  # TEMP, testing to see if negative data is messing everything up
    # limit the contrast by only considering data within a certain
    # range of the average
    if std_mult_cutoff is not None:
        avg = np.mean(temp_valid_data)
        std = np.std(temp_valid_data)
        # limit our range to avg +/- std_mult_cutoff*std; e.g. the
        # default std_mult_cutoff is 4.0 so about 99.8% of the data
        concervative_mask = (
                                    temp_valid_data < (avg + std * std_mult_cutoff)) & (
                                    temp_valid_data > (avg - std * std_mult_cutoff))
        temp_valid_data = temp_valid_data[concervative_mask]

    # if we are taking the log of our data, do so now
    if do_log_scale:
        temp_valid_data = np.log(temp_valid_data + log_offset)

    # do the histogram equalization and get the resulting
    # distribution function and bin information
    if not temp_valid_data.size:
        return None, None

    cumulative_dist_function, temp_bins = _histogram_equalization_helper(
        temp_valid_data,
        number_of_bins,
        clip_limit=clip_limit,
        slope_limit=slope_limit)
    return cumulative_dist_function, temp_bins


def _interpolate_local_equalized_tiles(
        data, out, mask_to_equalize, valid_data_mask, do_log_scale, log_offset,
        tile_weights, all_bin_information, all_cumulative_dist_functions,
        row_idx, col_idx, tile_size):
    # calculate the range for this tile (min is inclusive, max is exclusive)
    num_row_tile = row_idx
    num_col_tile = col_idx
    min_row = num_row_tile * tile_size
    max_row = min_row + tile_size
    min_col = num_col_tile * tile_size
    max_col = min_col + tile_size

    # for convenience, pull some of these tile sized chunks out
    temp_all_data = data[min_row:max_row, min_col:max_col].copy()
    temp_mask_to_equalize = mask_to_equalize[min_row:max_row, min_col:max_col]
    temp_all_valid_data_mask = valid_data_mask[min_row:max_row, min_col:max_col]

    # if we have any data in this tile, calculate our weighted sum
    if not temp_mask_to_equalize.any():
        return

    if do_log_scale:
        temp_all_data[temp_all_valid_data_mask] = np.log(
            temp_all_data[temp_all_valid_data_mask] + log_offset)
    temp_data_to_equalize = temp_all_data[temp_mask_to_equalize]
    temp_all_valid_data = temp_all_data[temp_all_valid_data_mask]

    # a place to hold our weighted sum that represents the interpolated contributions
    # of the histogram equalizations from the surrounding tiles
    temp_sum = np.zeros_like(temp_data_to_equalize)

    # how much weight were we unable to use because those tiles
    # fell off the edge of the image?
    unused_weight = np.zeros(temp_data_to_equalize.shape, dtype=tile_weights.dtype)

    # loop through all the surrounding tiles and process their
    # contributions to this tile
    for weight_row in range(3):
        for weight_col in range(3):
            tmp_tile_weights = tile_weights[weight_row, weight_col][np.where(temp_mask_to_equalize)]
            cumul_func, bin_info = _get_cumul_bin_info_for_tile(
                num_row_tile, weight_row,
                num_col_tile, weight_col,
                all_cumulative_dist_functions, all_bin_information,
            )
            if bin_info is None or cumul_func is None:
                unused_weight -= tmp_tile_weights
                continue

            # equalize our current tile using the histogram
            # equalization from the tile we're processing
            temp_equalized_data = np.interp(temp_all_valid_data, bin_info[:-1], cumul_func)
            temp_equalized_data = temp_equalized_data[np.where(
                temp_mask_to_equalize[temp_all_valid_data_mask])]

            # add the contribution for the tile we're
            # processing to our weighted sum
            temp_sum += temp_equalized_data * tmp_tile_weights

    # if we have unused weights, scale our values to correct for that
    if unused_weight.any():
        # TODO: if the mask masks everything out this will be a zero!
        temp_sum /= unused_weight + 1

    out[min_row:max_row, min_col:max_col][temp_mask_to_equalize] = temp_sum


def _get_cumul_bin_info_for_tile(
        num_row_tile, weight_row,
        num_col_tile, weight_col,
        all_cumulative_dist_functions, all_bin_information):
    # figure out which adjacent tile we're processing (in
    # overall tile coordinates instead of relative to our
    # current tile)
    calculated_row = num_row_tile - 1 + weight_row
    calculated_col = num_col_tile - 1 + weight_col
    if calculated_row < 0 or calculated_col < 0:
        # don't allow negative indexes (out of bounds)
        return None, None

    try:
        bin_info = all_bin_information[calculated_row][calculated_col]
        cumul_func = all_cumulative_dist_functions[calculated_row][calculated_col]
        return cumul_func, bin_info  # can be None
    except IndexError:
        return None, None


def _histogram_equalization_helper(valid_data, number_of_bins, clip_limit=None, slope_limit=None):
    """Calculate the simplest possible histogram equalization, using only valid data.

    Returns:
        cumulative distribution function and bin information

    """
    # bucket all the selected data using np's histogram function
    temp_histogram, temp_bins = np.histogram(valid_data, number_of_bins)

    # if we have a clip limit and we should do our clipping before building
    # the cumulative distribution function, clip off our histogram
    if clip_limit is not None:
        # clip our histogram and remember how much we removed
        pixels_to_clip_at = int(clip_limit *
                                (valid_data.size / float(number_of_bins)))
        mask_to_clip = temp_histogram > clip_limit
        # num_bins_clipped = sum(mask_to_clip)
        # num_pixels_clipped = sum(temp_histogram[mask_to_clip]) - (num_bins_clipped * pixels_to_clip_at)
        temp_histogram[mask_to_clip] = pixels_to_clip_at

    # calculate the cumulative distribution function
    cumulative_dist_function = temp_histogram.cumsum()

    # if we have a clip limit and we should do our clipping after building the
    # cumulative distribution function, clip off our cdf
    if slope_limit is not None:
        # clip our cdf and remember how much we removed
        pixel_height_limit = int(slope_limit *
                                 (valid_data.size / float(number_of_bins)))
        cumulative_excess_height = 0
        num_clipped_pixels = 0
        weight_metric = np.zeros(cumulative_dist_function.shape, dtype=float)

        for pixel_index in range(1, cumulative_dist_function.size):

            current_pixel_count = cumulative_dist_function[pixel_index]

            diff_from_acceptable = (
                current_pixel_count - cumulative_dist_function[pixel_index - 1]
                - pixel_height_limit - cumulative_excess_height)
            if diff_from_acceptable < 0:
                weight_metric[pixel_index] = abs(diff_from_acceptable)
            cumulative_excess_height += max(diff_from_acceptable, 0)
            cumulative_dist_function[
                pixel_index] = current_pixel_count - cumulative_excess_height
            num_clipped_pixels = num_clipped_pixels + cumulative_excess_height

    # now normalize the overall distribution function
    cumulative_dist_function = (number_of_bins - 1) * cumulative_dist_function / cumulative_dist_function[-1]

    # return what someone else will need in order to apply the equalization later
    return cumulative_dist_function, temp_bins


def _calculate_weights(tile_size):
    """Calculate a weight array for bilinear interpolation of histogram tiles.

    The weight array will be used to quickly bilinearly-interpolate the
    histogram equalizations tile size should be the width and height of a tile
    in pixels.

    Returns: 4D weight array where the first 2 dimensions correspond to the
        grid of where the tiles are relative to the tile being interpolated.
    """
    # we are essentially making a set of weight masks for an ideal center tile
    # that has all 8 surrounding tiles available
    # create our empty template tiles
    template_tile = np.zeros((3, 3, tile_size, tile_size), dtype=np.float32)
    # TEMP FOR TESTING, create a weight tile that does no interpolation
    # template_tile[1,1] = template_tile[1,1] + 1.0

    # for ease of calculation, figure out the index of the center pixel in a tile
    # and how far that pixel is from the edge of the tile (in pixel units)
    center_index = int(tile_size / 2)
    center_dist = tile_size / 2.0

    # loop through each pixel in the tile and calculate the 9 weights for that pixel
    # were weights for a pixel are 0.0 they are not set (since the template_tile
    # starts out as all zeros)
    for row in range(tile_size):
        for col in range(tile_size):

            vertical_dist = abs(
                center_dist - row
            )  # the distance from our pixel to the center of our tile, vertically
            horizontal_dist = abs(
                center_dist - col
            )  # the distance from our pixel to the center of our tile, horizontally

            # pre-calculate which 3 adjacent tiles will affect our tile
            # (note: these calculations aren't quite right if center_index equals the row or col)
            horizontal_index = 0 if col < center_index else 2
            vertical_index = 0 if row < center_index else 2

            # if this is the center pixel, we only need to use it's own tile
            # for it
            if (row is center_index) and (col is center_index):

                # all of the weight for this pixel comes from it's own tile
                template_tile[1, 1][row, col] = 1.0

            # if this pixel is in the center row, but is not the center pixel
            # we're going to need to linearly interpolate it's tile and the
            # tile that is horizontally nearest to it
            elif (row is center_index) and (col is not center_index):

                # linear interp horizontally

                beside_weight = horizontal_dist / tile_size  # the weight from the adjacent tile
                local_weight = (
                    tile_size -
                    horizontal_dist) / tile_size  # the weight from this tile

                # set the weights for the two relevant tiles
                template_tile[1, 1][row, col] = local_weight
                template_tile[1, horizontal_index][row, col] = beside_weight

            # if this pixel is in the center column, but is not the center pixel
            # we're going to need to linearly interpolate it's tile and the
            # tile that is vertically nearest to it
            elif (row is not center_index) and (col is center_index):

                # linear interp vertical

                beside_weight = vertical_dist / tile_size  # the weight from the adjacent tile
                local_weight = (
                    tile_size -
                    vertical_dist) / tile_size  # the weight from this tile

                # set the weights for the two relevant tiles
                template_tile[1, 1][row, col] = local_weight
                template_tile[vertical_index, 1][row, col] = beside_weight

            # if the pixel is in one of the four quadrants that are above or below the center
            # row and column, we need to bilinearly interpolate it between the
            # nearest four tiles
            else:

                # bilinear interpolation

                local_weight = ((tile_size - vertical_dist) / tile_size) * (
                    (tile_size - horizontal_dist) /
                    tile_size)  # the weight from this tile
                vertical_weight = ((vertical_dist) / tile_size) * (
                    (tile_size - horizontal_dist) / tile_size
                )  # the weight from the vertically   adjacent tile
                horizontal_weight = (
                    (tile_size - vertical_dist) / tile_size) * (
                        (horizontal_dist) / tile_size
                )  # the weight from the horizontally adjacent tile
                diagonal_weight = ((vertical_dist) / tile_size) * (
                    (horizontal_dist) / tile_size
                )  # the weight from the diagonally   adjacent tile

                # set the weights for the four relevant tiles
                template_tile[1, 1, row, col] = local_weight
                template_tile[vertical_index, 1, row, col] = vertical_weight
                template_tile[1, horizontal_index, row,
                              col] = horizontal_weight
                template_tile[vertical_index, horizontal_index, row,
                              col] = diagonal_weight

    # return the weights for an ideal center tile
    return template_tile


def _linear_normalization_from_0to1(
        data,
        mask,
        theoretical_max,
        theoretical_min=0,
        message="normalizing equalized data to fit in 0 to 1 range"):
    """Do a linear normalization so all data is in the 0 to 1 range.

    This is a sloppy but fast calculation that relies on parameters giving it
    the correct theoretical current max and min so it can scale the data
    accordingly.
    """
    LOG.debug(message)
    if theoretical_min != 0:
        data[mask] = data[mask] - theoretical_min
        theoretical_max = theoretical_max - theoretical_min
    data[mask] = data[mask] / theoretical_max


def _check_moon_phase(moon_datasets: list[xr.DataArray], start_time: dt.datetime) -> float:
    """Check if we have Moon phase as an input dataset and, if not, calculate it."""
    if moon_datasets:
        # convert to decimal instead of %
        return da.mean(moon_datasets[0].data) * 0.01

    LOG.debug("Moon illumination fraction not present. Calculating from start time.")
    try:
        import ephem
    except ImportError:
        raise ImportError("The 'ephem' library is required to calculate moon illumination fraction")
    return ephem.Moon(start_time).moon_phase


class NCCZinke(CompositeBase):
    """Equalized DNB composite using the Zinke algorithm [#ncc1]_.

    References:
        .. [#ncc1] Stephan Zinke (2017),

               A simplified high and near-constant contrast approach for the
               display of VIIRS day/night band imagery
               :doi:`10.1080/01431161.2017.1338838`

    """

    def __call__(self, datasets, **info):
        """Create HNCC DNB composite."""
        if len(datasets) < 3 or len(datasets) > 4:
            raise ValueError("Expected either 3 or 4 datasets, got %d" % (len(datasets),))
        dnb_data = datasets[0]
        sza_data = datasets[1]
        lza_data = datasets[2]
        moon_illum_fraction = _check_moon_phase(datasets[3:4], dnb_data.attrs["start_time"])

        # this algorithm assumes units of "W cm-2 sr-1" so if there are other
        # units we need to adjust for that
        if dnb_data.attrs.get("units", "W m-2 sr-1") == "W m-2 sr-1":
            unit_factor = 10000.
        else:
            unit_factor = 1.

        mda = dnb_data.attrs.copy()
        dnb_data = dnb_data.copy() / unit_factor

        phi = da.rad2deg(da.arccos(2. * moon_illum_fraction - 1))

        vfl = 0.026 * phi + 4.0e-9 * (phi ** 4.)

        m_fullmoon = -12.74
        m_sun = -26.74
        m_moon = vfl + m_fullmoon

        gs_ = self.gain_factor(sza_data.data)

        r_sun_moon = 10.**((m_sun - m_moon) / -2.5)
        gl_ = r_sun_moon * self.gain_factor(lza_data.data)
        gtot = 1. / (1. / gs_ + 1. / gl_)

        dnb_data += 2.6e-10
        dnb_data *= gtot

        mda["name"] = self.attrs["name"]
        mda["standard_name"] = "ncc_radiance"
        dnb_data.attrs = mda
        return dnb_data

    def gain_factor(self, theta):
        """Compute gain factor in a dask-friendly manner."""
        return theta.map_blocks(self._gain_factor, dtype=theta.dtype)

    @staticmethod
    def _gain_factor(theta):
        gain = np.empty_like(theta)

        mask = theta <= 87.541
        gain[mask] = (58 + 4 / np.cos(np.deg2rad(theta[mask]))) / 5

        mask = np.logical_and(theta <= 96, 87.541 < theta)
        gain[mask] = (123 * np.exp(1.06 * (theta[mask] - 89.589)) *
                      ((theta[mask] - 93)**2 / 18 + 0.5))

        mask = np.logical_and(96 < theta, theta <= 101)
        gain[mask] = 123 * np.exp(1.06 * (theta[mask] - 89.589))

        mask = np.logical_and(101 < theta, theta <= 103.49)
        gain[mask] = (123 * np.exp(1.06 * (101 - 89.589)) *
                      np.log(theta[mask] - (101 - np.e)) ** 2)

        gain[theta > 103.49] = 6.0e7

        return gain


class SnowAge(GenericCompositor):
    """Create RGB snow product.

    Product is based on method presented at the second
    CSPP/IMAPP users' meeting at Eumetsat in Darmstadt on 14-16 April 2015

    Bernard Bellec snow Look-Up Tables V 1.0 (c) Meteo-France
    These Look-up Tables allow you to create the RGB snow product
    for SUOMI-NPP VIIRS Imager according to the algorithm
    presented at the second CSPP/IMAPP users' meeting at Eumetsat
    in Darmstadt on 14-16 April 2015
    The algorithm and the product are described in this
    presentation :
    http://www.ssec.wisc.edu/meetings/cspp/2015/Agenda%20PDF/Wednesday/Roquet_snow_product_cspp2015.pdf
    as well as in the paper http://dx.doi.org/10.1016/j.rse.2017.04.028
    For further information you may contact
    Bernard Bellec at Bernard.Bellec@meteo.fr
    or
    Pascale Roquet at Pascale.Roquet@meteo.fr
    """

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate a SnowAge RGB composite.

        The algorithm and the product are described in this
        presentation :
        http://www.ssec.wisc.edu/meetings/cspp/2015/Agenda%20PDF/Wednesday/Roquet_snow_product_cspp2015.pdf
        as well as in the paper http://dx.doi.org/10.1016/j.rse.2017.04.028
        For further information you may contact
        Bernard Bellec at Bernard.Bellec@meteo.fr
        or
        Pascale Roquet at Pascale.Roquet@meteo.fr

        The resulting RGB has the units attribute removed.
        """
        if len(projectables) != 5:
            raise ValueError("Expected 5 datasets, got %d" %
                             (len(projectables), ))

        # Collect information that is the same between the projectables
        info = combine_metadata(*projectables)
        # Update that information with configured information (including name)
        info.update(self.attrs)
        # Force certain pieces of metadata that we *know* to be true
        info["wavelength"] = None

        m07 = projectables[0] * 255. / 160.
        m08 = projectables[1] * 255. / 160.
        m09 = projectables[2] * 255. / 160.
        m10 = projectables[3] * 255. / 160.
        m11 = projectables[4] * 255. / 160.
        refcu = m11 - m10
        refcu = refcu.clip(min=0)

        ch1 = m07 - refcu / 2. - m09 / 4.
        ch2 = m08 + refcu / 4. + m09 / 4.
        ch3 = m11 + m09
        # GenericCompositor needs valid DataArrays with 'area' metadata
        ch1.attrs = info
        ch2.attrs = info
        ch3.attrs = info

        res = super(SnowAge, self).__call__([ch1, ch2, ch3], **info)
        res.attrs.pop("units", None)
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Shared preparation and utilities for testing.

This module is executed automatically by pytest.

"""
import os

import pytest

import satpy

TEST_ETC_DIR = os.path.join(os.path.dirname(__file__), "etc")


@pytest.fixture(autouse=True)
def _reset_satpy_config(tmpdir):
    """Set satpy config to logical defaults for tests."""
    test_config = {
        "cache_dir": str(tmpdir / "cache"),
        "data_dir": str(tmpdir / "data"),
        "config_path": [],
        "cache_lonlats": False,
        "cache_sensor_angles": False,
    }
    with satpy.config.set(test_config):
        yield


@pytest.fixture(autouse=True)
def _clear_function_caches():
    """Clear out global function-level caches that may cause conflicts between tests."""
    from satpy.composites.config_loader import load_compositor_configs_for_sensor
    load_compositor_configs_for_sensor.cache_clear()


@pytest.fixture
def include_test_etc():
    """Tell Satpy to use the config 'etc' directory from the tests directory."""
    with satpy.config.set(config_path=[TEST_ETC_DIR]):
        yield TEST_ETC_DIR


@pytest.fixture(autouse=True, scope="session")
def _forbid_pyspectral_downloads():
    from pyspectral.testing import forbid_pyspectral_downloads

    with forbid_pyspectral_downloads():
        yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""test file handler baseclass."""

import datetime as dt
import unittest
from unittest import mock

import numpy as np
import pytest

from satpy.readers.core.file_handlers import BaseFileHandler, open_dataset
from satpy.tests.utils import FakeFileHandler


def test_open_dataset():
    """Test xr.open_dataset wrapper."""
    fn = mock.MagicMock()
    str_file_path = "path/to/file.nc"
    with mock.patch("xarray.open_dataset") as xr_open:
        _ = open_dataset(fn, decode_cf=True, chunks=500)
        fn.open.assert_called_once_with()
        xr_open.assert_called_once_with(fn.open(), decode_cf=True, chunks=500)

        xr_open.reset_mock()
        _ = open_dataset(str_file_path, decode_cf=True, chunks=500)
        xr_open.assert_called_once_with(str_file_path, decode_cf=True, chunks=500)


class TestBaseFileHandler(unittest.TestCase):
    """Test the BaseFileHandler."""

    def setUp(self):
        """Set up the test."""
        self.fh = BaseFileHandler(
            "filename", {"filename_info": "bla"}, "filetype_info")
        self.early_time = dt.datetime(2024, 2, 12, 11, 00)
        self.late_time = dt.datetime(2024, 2, 12, 12, 00)

    def test_combine_times(self):
        """Combine times."""
        info1 = {"start_time": self.early_time}
        info2 = {"start_time": self.late_time}
        res = self.fh.combine_info([info1, info2])
        exp = {"start_time": self.early_time}
        assert res == exp
        res = self.fh.combine_info([info2, info1])
        exp = {"start_time": self.early_time}
        assert res == exp

        info1 = {"end_time": self.early_time}
        info2 = {"end_time": self.late_time}
        res = self.fh.combine_info([info1, info2])
        exp = {"end_time": self.late_time}
        assert res == exp
        res = self.fh.combine_info([info2, info1])
        exp = {"end_time": self.late_time}
        assert res == exp

    def test_combine_orbits(self):
        """Combine orbits."""
        info1 = {"start_orbit": 1}
        info2 = {"start_orbit": 2}
        res = self.fh.combine_info([info1, info2])
        exp = {"start_orbit": 1}
        assert res == exp
        res = self.fh.combine_info([info2, info1])
        exp = {"start_orbit": 1}
        assert res == exp

        info1 = {"end_orbit": 1}
        info2 = {"end_orbit": 2}
        res = self.fh.combine_info([info1, info2])
        exp = {"end_orbit": 2}
        assert res == exp
        res = self.fh.combine_info([info2, info1])
        exp = {"end_orbit": 2}
        assert res == exp

    @mock.patch("satpy.readers.core.file_handlers.SwathDefinition")
    def test_combine_area(self, sdef):
        """Combine area."""
        area1 = mock.MagicMock()
        area1.lons = np.arange(5)
        area1.lats = np.arange(5)
        area1.name = "area1"

        area2 = mock.MagicMock()
        area2.lons = np.arange(5)
        area2.lats = np.arange(5)
        area2.name = "area2"

        info1 = {"area": area1}
        info2 = {"area": area2}

        self.fh.combine_info([info1, info2])
        assert sdef.call_args[1]["lons"].shape == (2, 5)
        assert sdef.call_args[1]["lats"].shape == (2, 5)
        assert sdef.return_value.name == "area1_area2"

    def test_combine_orbital_parameters(self):
        """Combine orbital parameters."""
        info1 = {"orbital_parameters": {"projection_longitude": 1,
                                        "projection_latitude": 1,
                                        "projection_altitude": 1,
                                        "satellite_nominal_longitude": 1,
                                        "satellite_nominal_latitude": 1,
                                        "satellite_actual_longitude": 1,
                                        "satellite_actual_latitude": 1,
                                        "satellite_actual_altitude": 1,
                                        "nadir_longitude": 1,
                                        "nadir_latitude": 1,
                                        "only_in_1": False}}
        info2 = {"orbital_parameters": {"projection_longitude": 2,
                                        "projection_latitude": 2,
                                        "projection_altitude": 2,
                                        "satellite_nominal_longitude": 2,
                                        "satellite_nominal_latitude": 2,
                                        "satellite_actual_longitude": 2,
                                        "satellite_actual_latitude": 2,
                                        "satellite_actual_altitude": 2,
                                        "nadir_longitude": 2,
                                        "nadir_latitude": 2,
                                        "only_in_2": True}}
        exp = {"orbital_parameters": {"projection_longitude": 1.5,
                                      "projection_latitude": 1.5,
                                      "projection_altitude": 1.5,
                                      "satellite_nominal_longitude": 1.5,
                                      "satellite_nominal_latitude": 1.5,
                                      "satellite_actual_longitude": 1.5,
                                      "satellite_actual_latitude": 1.5,
                                      "satellite_actual_altitude": 1.5,
                                      "nadir_longitude": 1.5,
                                      "nadir_latitude": 1.5,
                                      "only_in_1": False,
                                      "only_in_2": True}}
        res = self.fh.combine_info([info1, info2])
        assert res == exp

        # Identity
        assert self.fh.combine_info([info1]) == info1

        # Empty
        self.fh.combine_info([{}])

    def test_combine_time_parameters(self):
        """Combine times in 'time_parameters."""
        time_params1 = {
            "nominal_start_time": dt.datetime(2020, 1, 1, 12, 0, 0),
            "nominal_end_time": dt.datetime(2020, 1, 1, 12, 2, 30),
            "observation_start_time": dt.datetime(2020, 1, 1, 12, 0, 2, 23821),
            "observation_end_time": dt.datetime(2020, 1, 1, 12, 2, 23, 12348),
        }
        time_params2 = {}
        time_shift = dt.timedelta(seconds=1.5)
        for key, value in time_params1.items():
            time_params2[key] = value + time_shift
        res = self.fh.combine_info([
            {"time_parameters": time_params1},
            {"time_parameters": time_params2}
        ])
        res_time_params = res["time_parameters"]
        assert res_time_params["nominal_start_time"] == dt.datetime(2020, 1, 1, 12, 0, 0)
        assert res_time_params["nominal_end_time"] == dt.datetime(2020, 1, 1, 12, 2, 31, 500000)
        assert res_time_params["observation_start_time"] == dt.datetime(2020, 1, 1, 12, 0, 2, 23821)
        assert res_time_params["observation_end_time"] == dt.datetime(2020, 1, 1, 12, 2, 24, 512348)

    def test_file_is_kept_intact(self):
        """Test that the file object passed (string, path, or other) is kept intact."""
        open_file = mock.MagicMock()
        bfh = BaseFileHandler(open_file, {"filename_info": "bla"}, "filetype_info")
        assert bfh.filename == open_file

        from pathlib import Path
        filename = Path("/bla/bla.nc")
        bfh = BaseFileHandler(filename, {"filename_info": "bla"}, "filetype_info")
        assert isinstance(bfh.filename, Path)


@pytest.mark.parametrize(
    ("file_type", "ds_file_type", "exp_result"),
    [
        ("fake1", "fake1", True),
        ("fake1", ["fake1"], True),
        ("fake1", ["fake1", "fake2"], True),
        ("fake1", ["fake2"], None),
        ("fake1", "fake2", None),
        ("fake1", "fake1_with_suffix", None),
    ]
)
def test_file_type_match(file_type, ds_file_type, exp_result):
    """Test that file type matching uses exactly equality."""
    fh = FakeFileHandler("some_file.txt", {}, {"file_type": file_type})
    assert fh.file_type_matches(ds_file_type) is exp_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, 2022, 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test classes and functions in the readers/__init__.py module."""

import builtins
import contextlib
import datetime as dt
import os
import sys
import unittest
import warnings
from pathlib import Path
from typing import Iterator
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy.dataset.data_dict import get_key
from satpy.dataset.dataid import DataID, ModifierTuple, WavelengthRange
from satpy.readers.core.grouping import find_files_and_readers
from satpy.readers.core.remote import open_file_or_filename

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - monkeypatch
# - tmp_path

# clear the config dir environment variable so it doesn't interfere
os.environ.pop("PPP_CONFIG_DIR", None)
os.environ.pop("SATPY_CONFIG_PATH", None)

local_id_keys_config = {"name": {
    "required": True,
},
    "wavelength": {
    "type": WavelengthRange,
},
    "resolution": None,
    "calibration": {
    "enum": [
        "reflectance",
        "brightness_temperature",
        "radiance",
        "counts"
    ]
},
    "polarization": None,
    "level": None,
    "modifiers": {
    "required": True,
    "default": ModifierTuple(),
    "type": ModifierTuple,
},
}

real_import = builtins.__import__


@pytest.fixture
def viirs_file(tmp_path, monkeypatch):
    """Create a dummy viirs file."""
    filename = "SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"

    monkeypatch.chdir(tmp_path)
    # touch the file so it exists on disk
    open(filename, "w").close()
    return filename


@pytest.fixture
def atms_file(tmp_path, monkeypatch):
    """Create a dummy atms file."""
    filename = "SATMS_j01_d20221220_t0910240_e0921356_b26361_c20221220100456348770_cspp_dev.h5"

    monkeypatch.chdir(tmp_path)
    # touch the file so it exists on disk
    open(filename, "w").close()
    return filename


def make_dataid(**items):
    """Make a data id."""
    return DataID(local_id_keys_config, **items)


class TestDatasetDict(unittest.TestCase):
    """Test DatasetDict and its methods."""

    def setUp(self):
        """Create a test DatasetDict."""
        from satpy import DatasetDict
        self.regular_dict = regular_dict = {
            make_dataid(name="test",
                        wavelength=(0, 0.5, 1),
                        resolution=1000): "1",
            make_dataid(name="testh",
                        wavelength=(0, 0.5, 1),
                        resolution=500): "1h",
            make_dataid(name="test2",
                        wavelength=(1, 1.5, 2),
                        resolution=1000): "2",
            make_dataid(name="test3",
                        wavelength=(1.2, 1.7, 2.2),
                        resolution=1000): "3",
            make_dataid(name="test4",
                        calibration="radiance",
                        polarization="V"): "4rad",
            make_dataid(name="test4",
                        calibration="reflectance",
                        polarization="H"): "4refl",
            make_dataid(name="test5",
                        modifiers=("mod1", "mod2")): "5_2mod",
            make_dataid(name="test5",
                        modifiers=("mod2",)): "5_1mod",
            make_dataid(name="test6", level=100): "6_100",
            make_dataid(name="test6", level=200): "6_200",
        }
        self.test_dict = DatasetDict(regular_dict)

    def test_init_noargs(self):
        """Test DatasetDict init with no arguments."""
        from satpy import DatasetDict
        d = DatasetDict()
        assert isinstance(d, dict)

    def test_init_dict(self):
        """Test DatasetDict init with a regular dict argument."""
        from satpy import DatasetDict
        regular_dict = {make_dataid(name="test", wavelength=(0, 0.5, 1)): "1", }
        d = DatasetDict(regular_dict)
        assert d == regular_dict

    def test_getitem(self):
        """Test DatasetDict getitem with different arguments."""
        from satpy.tests.utils import make_dsq
        d = self.test_dict
        # access by name
        assert d["test"] == "1"
        # access by exact wavelength
        assert d[1.5] == "2"
        # access by near wavelength
        assert d[1.55] == "2"
        # access by near wavelength of another dataset
        assert d[1.65] == "3"
        # access by name with multiple levels
        assert d["test6"] == "6_100"

        assert d[make_dsq(wavelength=1.5)] == "2"
        assert d[make_dsq(wavelength=0.5, resolution=1000)] == "1"
        assert d[make_dsq(wavelength=0.5, resolution=500)] == "1h"
        assert d[make_dsq(name="test6", level=100)] == "6_100"
        assert d[make_dsq(name="test6", level=200)] == "6_200"

        # higher resolution is returned
        assert d[0.5] == "1h"
        assert d["test4"] == "4refl"
        assert d[make_dataid(name="test4", calibration="radiance")] == "4rad"
        with pytest.raises(KeyError):
            d.getitem("1h")

        # test with full tuple
        assert d[make_dsq(name="test", wavelength=(0, 0.5, 1), resolution=1000)] == "1"

    def test_get_key(self):
        """Test 'get_key' special functions."""
        from satpy.dataset import DataQuery
        d = self.test_dict
        res1 = get_key(make_dataid(name="test4"), d, calibration="radiance")
        res2 = get_key(make_dataid(name="test4"), d, calibration="radiance",
                       num_results=0)
        res3 = get_key(make_dataid(name="test4"), d, calibration="radiance",
                       num_results=3)
        assert len(res2) == 1
        assert len(res3) == 1
        res2 = res2[0]
        res3 = res3[0]
        assert res1 == res2
        assert res1 == res3
        res1 = get_key("test4", d, query=DataQuery(polarization="V"))
        assert res1 == make_dataid(name="test4", calibration="radiance", polarization="V")

        res1 = get_key(0.5, d, query=DataQuery(resolution=500))
        assert res1 == make_dataid(name="testh", wavelength=(0, 0.5, 1), resolution=500)

        res1 = get_key("test6", d, query=DataQuery(level=100))
        assert res1 == make_dataid(name="test6", level=100)

        res1 = get_key("test5", d)
        res2 = get_key("test5", d, query=DataQuery(modifiers=("mod2",)))
        res3 = get_key("test5", d, query=DataQuery(modifiers=("mod1", "mod2",)))
        assert res1 == make_dataid(name="test5", modifiers=("mod2",))
        assert res1 == res2
        assert res1 != res3

        # more than 1 result when default is to ask for 1 result
        with pytest.raises(KeyError):
            get_key("test4", d, best=False)

    def test_contains(self):
        """Test DatasetDict contains method."""
        d = self.test_dict
        assert "test" in d
        assert not d.contains("test")
        assert "test_bad" not in d
        assert 0.5 in d
        assert not d.contains(0.5)
        assert 1.5 in d
        assert 1.55 in d
        assert 1.65 in d
        assert make_dataid(name="test4", calibration="radiance") in d
        assert "test4" in d

    def test_keys(self):
        """Test keys method of DatasetDict."""
        from satpy.tests.utils import DataID
        d = self.test_dict
        assert len(d.keys()) == len(self.regular_dict.keys())
        assert all(isinstance(x, DataID) for x in d.keys())
        name_keys = d.keys(names=True)
        assert sorted(set(name_keys))[:4] == ["test", "test2", "test3", "test4"]
        wl_keys = tuple(d.keys(wavelengths=True))
        assert (0, 0.5, 1) in wl_keys
        assert (1, 1.5, 2, "µm") in wl_keys
        assert (1.2, 1.7, 2.2, "µm") in wl_keys
        assert None in wl_keys

    def test_setitem(self):
        """Test setitem method of DatasetDict."""
        d = self.test_dict
        d["new_ds"] = {"metadata": "new_ds"}
        assert d["new_ds"]["metadata"] == "new_ds"
        d[0.5] = {"calibration": "radiance"}
        assert d[0.5]["resolution"] == 500
        assert d[0.5]["name"] == "testh"


class TestReaderLoader(unittest.TestCase):
    """Test the `load_readers` function.

    Assumes that the VIIRS SDR reader exists and works.
    """

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog, tmp_path):  # noqa: PT004
        """Inject caplog to the test class."""
        self._caplog = caplog
        self._tmp_path = tmp_path

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy.readers.viirs_sdr import VIIRSSDRFileHandler
        from satpy.tests.reader_tests.test_viirs_sdr import FakeHDF5FileHandler2

        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIIRSSDRFileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_no_args(self):
        """Test no args provided.

        This should check the local directory which should have no files.
        """
        from satpy.readers.core.loading import load_readers
        ri = load_readers()
        assert ri == {}

    def test_filenames_only(self):
        """Test with filenames specified."""
        from satpy.readers.core.loading import load_readers
        ri = load_readers(filenames=["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"])
        assert list(ri.keys()) == ["viirs_sdr"]

    def test_filenames_and_reader(self):
        """Test with filenames and reader specified."""
        from satpy.readers.core.loading import load_readers
        ri = load_readers(reader="viirs_sdr",
                          filenames=["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"])
        assert list(ri.keys()) == ["viirs_sdr"]

    def test_bad_reader_name_with_filenames(self):
        """Test bad reader name with filenames provided."""
        from satpy.readers.core.loading import load_readers
        with pytest.raises(ValueError, match="No reader named: i_dont_exist"):
            load_readers(reader="i_dont_exist",
                         filenames=["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"])

    def test_filenames_as_path(self):
        """Test with filenames specified as pathlib.Path."""
        from pathlib import Path

        from satpy.readers.core.loading import load_readers
        ri = load_readers(filenames=[
            Path("SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"),
        ])
        assert list(ri.keys()) == ["viirs_sdr"]

    def test_filenames_as_dict(self):
        """Test loading readers where filenames are organized by reader."""
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
        }
        ri = load_readers(filenames=filenames)
        assert list(ri.keys()) == ["viirs_sdr"]

    def test_filenames_as_dict_bad_reader(self):
        """Test loading with filenames dict but one of the readers is bad."""
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
            "__fake__": ["fake.txt"],
        }
        with pytest.raises(ValueError, match=r"(?=.*__fake__)(?!.*viirs)(^No reader.+)"):
            load_readers(filenames=filenames)

    def test_filenames_as_dict_with_reader(self):
        """Test loading from a filenames dict with a single reader specified.

        This can happen in the deprecated Scene behavior of passing a reader
        and a base_dir.

        """
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
        }
        ri = load_readers(reader="viirs_sdr", filenames=filenames)
        assert list(ri.keys()) == ["viirs_sdr"]

    def test_empty_filenames_as_dict(self):
        """Test passing filenames as a dictionary with an empty list of filenames."""
        # only one reader
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": [],
        }
        with pytest.raises(ValueError, match="No supported files found"):
            load_readers(filenames=filenames)

        # two readers, one is empty
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
            "viirs_l1b": [],
        }
        ri = load_readers(filenames)
        assert list(ri.keys()) == ["viirs_sdr"]

    @mock.patch("satpy.readers.core.hrit.HRITFileHandler._get_hd")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler._get_header")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.start_time")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.end_time")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGPrologueFileHandler.read_prologue")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGEpilogueFileHandler.read_epilogue")
    def test_missing_requirements(self, *mocks):
        """Test warnings and exceptions in case of missing requirements."""
        from satpy.readers.core.loading import load_readers

        # Filenames from a single scan
        epi_pro_miss = ["H-000-MSG4__-MSG4________-IR_108___-000006___-201809050900-__"]
        epi_miss = epi_pro_miss + ["H-000-MSG4__-MSG4________-_________-PRO______-201809050900-__"]
        pro_miss = epi_pro_miss + ["H-000-MSG4__-MSG4________-_________-EPI______-201809050900-__"]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=r"No handler for reading requirement.*", category=UserWarning)
            for filenames in [epi_miss, pro_miss, epi_pro_miss]:
                with pytest.raises(ValueError, match="No dataset could be loaded.*"):
                    load_readers(reader="seviri_l1b_hrit", filenames=filenames)

        # Filenames from multiple scans
        at_least_one_complete = [
            # 09:00 scan is ok
            "H-000-MSG4__-MSG4________-IR_108___-000006___-201809050900-__",
            "H-000-MSG4__-MSG4________-_________-PRO______-201809050900-__",
            "H-000-MSG4__-MSG4________-_________-EPI______-201809050900-__",
            # 10:00 scan is incomplete
            "H-000-MSG4__-MSG4________-IR_108___-000006___-201809051000-__",
        ]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=r"No matching requirement file.*", category=UserWarning)
            try:
                load_readers(filenames=at_least_one_complete, reader="seviri_l1b_hrit")
            except ValueError:
                self.fail("If at least one set of filenames is complete, no "
                          "exception should be raised")

    def test_all_filtered(self):
        """Test behaviour if no file matches the filter parameters."""
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
        }
        filter_params = {"start_time": dt.datetime(1970, 1, 1),
                         "end_time": dt.datetime(1970, 1, 2),
                         "area": None}
        with pytest.raises(ValueError, match="No dataset could be loaded.*"):
            load_readers(filenames=filenames, reader_kwargs={"filter_parameters": filter_params})

    def test_all_filtered_multiple(self):
        """Test behaviour if no file matches the filter parameters."""
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
            "abi_l1b": ["OR_ABI-L1b-RadF-M3C01_G16_s20120561730408_e20120561741175_c20172631741218.nc"],
        }
        filter_params = {"start_time": dt.datetime(1970, 1, 1),
                         "end_time": dt.datetime(1970, 1, 2)}
        with pytest.raises(ValueError, match="No dataset could be loaded."):
            load_readers(filenames=filenames, reader_kwargs={"filter_parameters": filter_params})

    def test_almost_all_filtered(self):
        """Test behaviour if only one reader has datasets."""
        from satpy.readers.core.loading import load_readers
        filenames = {
            "viirs_sdr": ["SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5"],
            "abi_l1b": ["OR_ABI-L1b-RadF-M3C01_G16_s20172631730408_e20172631741175_c20172631741218.nc"],
        }
        filter_params = {"start_time": dt.datetime(2012, 2, 25),
                         "end_time": dt.datetime(2012, 2, 26)}
        # viirs has data that matches the request, abi doesn't
        readers = load_readers(filenames=filenames, reader_kwargs={"filter_parameters": filter_params})
        assert "viirs_sdr" in readers
        # abi_l1b reader was created, but no datasets available
        assert "abi_l1b" in readers
        assert len(list(readers["abi_l1b"].available_dataset_ids)) == 0

    def test_yaml_error_message(self):
        """Test that YAML errors are logged properly."""
        import logging

        import satpy
        from satpy.readers.core.loading import load_readers

        reader_config = "reader:\n"
        reader_config += "  name: nonreader\n"
        reader_config += "  reader: !!python/name:notapackage.notareader.BadClass\n"

        os.mkdir(self._tmp_path / "readers")
        reader_fname = self._tmp_path / "readers" / "nonreader.yaml"
        with open(reader_fname, "w") as fid:
            fid.write(reader_config)

        filenames = ["foo.bar"]
        error_message = "No module named 'notapackage'"

        with self._caplog.at_level(logging.ERROR):
            with satpy.config.set({"config_path": [str(self._tmp_path)]}):
                with pytest.raises(ValueError, match="No supported files found"):
                    _ = load_readers(filenames=filenames, reader="nonreader")
            assert error_message in self._caplog.text


class TestFindFilesAndReaders:
    """Test the find_files_and_readers utility function."""

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy.readers.viirs_sdr import VIIRSSDRFileHandler
        from satpy.tests.reader_tests.test_viirs_sdr import FakeHDF5FileHandler2

        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIIRSSDRFileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_reader_name(self, viirs_file):
        """Test with default base_dir and reader specified."""
        ri = find_files_and_readers(reader="viirs_sdr")
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_reader_other_name(self, monkeypatch, tmp_path):
        """Test with default base_dir and reader specified."""
        filename = "S_NWC_CPP_npp_32505_20180204T1114116Z_20180204T1128227Z.nc"
        monkeypatch.chdir(tmp_path)
        # touch the file so it exists on disk
        open(filename, "w").close()

        ri = find_files_and_readers(reader="nwcsaf-pps_nc")
        assert list(ri.keys()) == ["nwcsaf-pps_nc"]
        assert ri["nwcsaf-pps_nc"] == [filename]

    def test_reader_name_matched_start_end_time(self, viirs_file):
        """Test with start and end time matching the filename."""
        ri = find_files_and_readers(reader="viirs_sdr",
                                    start_time=dt.datetime(2012, 2, 25, 18, 0, 0),
                                    end_time=dt.datetime(2012, 2, 25, 19, 0, 0),
                                    )
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_reader_name_matched_start_time(self, viirs_file):
        """Test with start matching the filename.

        Start time in the middle of the file time should still match the file.
        """
        ri = find_files_and_readers(reader="viirs_sdr", start_time=dt.datetime(2012, 2, 25, 18, 1, 30))
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_reader_name_matched_end_time(self, viirs_file):
        """Test with end matching the filename.

        End time in the middle of the file time should still match the file.

        """
        ri = find_files_and_readers(reader="viirs_sdr", end_time=dt.datetime(2012, 2, 25, 18, 1, 30))
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_reader_name_unmatched_start_end_time(self, viirs_file):
        """Test with start and end time matching the filename."""
        with pytest.raises(ValueError, match="No supported files found"):
            find_files_and_readers(reader="viirs_sdr",
                                   start_time=dt.datetime(2012, 2, 26, 18, 0, 0),
                                   end_time=dt.datetime(2012, 2, 26, 19, 0, 0))

    def test_no_parameters(self, viirs_file):
        """Test with no limiting parameters."""
        from satpy.readers.core.grouping import find_files_and_readers

        ri = find_files_and_readers()
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_no_parameters_both_atms_and_viirs(self, viirs_file, atms_file):
        """Test with no limiting parameters when there area both atms and viirs files in the same directory."""
        from satpy.readers.core.grouping import find_files_and_readers

        ri = find_files_and_readers()

        assert "atms_sdr_hdf5" in list(ri.keys())
        assert "viirs_sdr" in list(ri.keys())
        assert ri["atms_sdr_hdf5"] == [atms_file]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_bad_sensor(self):
        """Test bad sensor doesn't find any files."""
        with pytest.raises(ValueError, match="Sensor.* not supported by any readers"):
            find_files_and_readers(sensor="i_dont_exist")

    def test_sensor(self, viirs_file):
        """Test that readers for the current sensor are loaded."""
        # we can't easily know how many readers satpy has that support
        # 'viirs' so we just pass it and hope that this works
        ri = find_files_and_readers(sensor="viirs")
        assert list(ri.keys()) == ["viirs_sdr"]
        assert ri["viirs_sdr"] == [viirs_file]

    def test_sensor_no_files(self, tmp_path, monkeypatch):
        """Test that readers for the current sensor are loaded."""
        monkeypatch.chdir(tmp_path)
        # we can't easily know how many readers satpy has that support
        # 'viirs' so we just pass it and hope that this works
        with pytest.raises(ValueError, match="No supported files found"):
            find_files_and_readers(sensor="viirs")
        assert find_files_and_readers(sensor="viirs", missing_ok=True) == {}

    def test_reader_load_failed(self):
        """Test that an exception is raised when a reader can't be loaded."""
        import yaml

        from satpy.readers.core.grouping import find_files_and_readers

        # touch the file so it exists on disk
        with mock.patch("yaml.load") as load:
            load.side_effect = yaml.YAMLError("Import problems")
            with pytest.raises(yaml.YAMLError):
                find_files_and_readers(reader="viirs_sdr")

    def test_pending_old_reader_name_mapping(self):
        """Test that requesting pending old reader names raises a warning."""
        from satpy.readers.core.config import PENDING_OLD_READER_NAMES, get_valid_reader_names
        if not PENDING_OLD_READER_NAMES:
            return unittest.skip("Skipping pending deprecated reader tests because "
                                 "no pending deprecated readers.")
        test_reader = sorted(PENDING_OLD_READER_NAMES.keys())[0]
        with pytest.warns(FutureWarning, match=".*will be removed.*"):
            valid_reader_names = get_valid_reader_names([test_reader])
        assert valid_reader_names[0] == PENDING_OLD_READER_NAMES[test_reader]

    def test_old_reader_name_mapping(self):
        """Test that requesting old reader names raises a warning."""
        from satpy.readers.core.config import OLD_READER_NAMES, get_valid_reader_names
        if not OLD_READER_NAMES:
            return pytest.skip("Skipping deprecated reader tests because "
                               "no deprecated readers.")
        test_reader = sorted(OLD_READER_NAMES.keys())[0]
        with pytest.raises(ValueError, match="Reader name .* has been deprecated, use .* instead."):
            get_valid_reader_names([test_reader])


class TestYAMLFiles:
    """Test and analyze the reader configuration files."""

    def test_filename_matches_reader_name(self):
        """Test that every reader filename matches the name in the YAML."""
        import yaml

        class IgnoreLoader(yaml.SafeLoader):
            def _ignore_all_tags(self, tag_suffix, node):
                return tag_suffix + " " + node.value
        IgnoreLoader.add_multi_constructor("", IgnoreLoader._ignore_all_tags)

        from satpy._config import glob_config
        from satpy.readers.core.config import read_reader_config
        for reader_config in glob_config("readers/*.yaml"):
            reader_fn = os.path.basename(reader_config)
            reader_fn_name = os.path.splitext(reader_fn)[0]
            reader_info = read_reader_config([reader_config],
                                             loader=IgnoreLoader)
            assert reader_fn_name == reader_info["name"], \
                "Reader YAML filename doesn't match reader name in the YAML file."

    def test_available_readers(self):
        """Test the 'available_readers' function."""
        from satpy import available_readers

        reader_names = available_readers()
        assert len(reader_names) > 0
        assert isinstance(reader_names[0], str)
        assert "viirs_sdr" in reader_names  # needs h5py
        assert "abi_l1b" in reader_names  # needs netcdf4
        assert reader_names == sorted(reader_names)

        reader_infos = available_readers(as_dict=True)
        assert len(reader_names) == len(reader_infos)
        assert isinstance(reader_infos[0], dict)
        for reader_info in reader_infos:
            assert "name" in reader_info
        assert reader_infos == sorted(reader_infos, key=lambda reader_info: reader_info["name"])

    def test_available_readers_base_loader(self, monkeypatch):
        """Test the 'available_readers' function for yaml loader type BaseLoader."""
        import yaml

        from satpy import available_readers
        from satpy._config import glob_config

        def patched_import_error(name, globals=None, locals=None, fromlist=(), level=0):  # noqa: A002
            if name in ("netcdf4", ):
                raise ImportError(f"Mocked import error {name}")
            return real_import(name, globals=globals, locals=locals, fromlist=fromlist, level=level)

        monkeypatch.delitem(sys.modules, "netcdf4", raising=False)
        monkeypatch.setattr(builtins, "__import__", patched_import_error)

        with pytest.raises(ImportError):
            import netcdf4  # noqa: F401

        reader_names = available_readers(yaml_loader=yaml.BaseLoader)
        assert "abi_l1b" in reader_names  # needs netcdf4
        assert "viirs_l1b" in reader_names
        assert len(reader_names) == len(list(glob_config("readers/*.yaml")))


class TestGroupFiles(unittest.TestCase):
    """Test the 'group_files' utility function."""

    def setUp(self):
        """Set up test filenames to use."""
        input_files = [
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171502203_e20171171504576_c20171171505018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171507203_e20171171509576_c20171171510018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171512203_e20171171514576_c20171171515017.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171517203_e20171171519577_c20171171520019.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171522203_e20171171524576_c20171171525020.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171527203_e20171171529576_c20171171530017.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171502203_e20171171504576_c20171171505008.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171507203_e20171171509576_c20171171510012.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171512203_e20171171514576_c20171171515007.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171517203_e20171171519576_c20171171520010.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171522203_e20171171524576_c20171171525008.nc",
            "OR_ABI-L1b-RadC-M3C02_G16_s20171171527203_e20171171529576_c20171171530008.nc",
        ]
        self.g16_files = input_files
        self.g17_files = [x.replace("G16", "G17") for x in input_files]
        self.noaa20_files = [
            "GITCO_j01_d20180511_t2027292_e2028538_b02476_c20190530192858056873_noac_ops.h5",
            "GITCO_j01_d20180511_t2028550_e2030195_b02476_c20190530192932937427_noac_ops.h5",
            "GITCO_j01_d20180511_t2030208_e2031435_b02476_c20190530192932937427_noac_ops.h5",
            "GITCO_j01_d20180511_t2031447_e2033092_b02476_c20190530192932937427_noac_ops.h5",
            "GITCO_j01_d20180511_t2033105_e2034350_b02476_c20190530192932937427_noac_ops.h5",
            "SVI03_j01_d20180511_t2027292_e2028538_b02476_c20190530190950789763_noac_ops.h5",
            "SVI03_j01_d20180511_t2028550_e2030195_b02476_c20190530192911205765_noac_ops.h5",
            "SVI03_j01_d20180511_t2030208_e2031435_b02476_c20190530192911205765_noac_ops.h5",
            "SVI03_j01_d20180511_t2031447_e2033092_b02476_c20190530192911205765_noac_ops.h5",
            "SVI03_j01_d20180511_t2033105_e2034350_b02476_c20190530192911205765_noac_ops.h5",
            "SVI04_j01_d20180511_t2027292_e2028538_b02476_c20190530190951848958_noac_ops.h5",
            "SVI04_j01_d20180511_t2028550_e2030195_b02476_c20190530192903985164_noac_ops.h5",
            "SVI04_j01_d20180511_t2030208_e2031435_b02476_c20190530192903985164_noac_ops.h5",
            "SVI04_j01_d20180511_t2031447_e2033092_b02476_c20190530192903985164_noac_ops.h5",
            "SVI04_j01_d20180511_t2033105_e2034350_b02476_c20190530192903985164_noac_ops.h5"
        ]
        self.npp_files = [
            "GITCO_npp_d20180511_t1939067_e1940309_b33872_c20190612031740518143_noac_ops.h5",
            "GITCO_npp_d20180511_t1940321_e1941563_b33872_c20190612031740518143_noac_ops.h5",
            "GITCO_npp_d20180511_t1941575_e1943217_b33872_c20190612031740518143_noac_ops.h5",
            "SVI03_npp_d20180511_t1939067_e1940309_b33872_c20190612032009230105_noac_ops.h5",
            "SVI03_npp_d20180511_t1940321_e1941563_b33872_c20190612032009230105_noac_ops.h5",
            "SVI03_npp_d20180511_t1941575_e1943217_b33872_c20190612032009230105_noac_ops.h5",
        ]
        self.unknown_files = [
            "ʌsɔ˙pıʃɐʌuı",
            "no such"]

    def test_no_reader(self):
        """Test that reader does not need to be provided."""
        from satpy.readers.core.grouping import group_files

        # without files it's going to be an empty result
        assert group_files([]) == []
        groups = group_files(self.g16_files)
        assert 6 == len(groups)

    def test_unknown_files(self):
        """Test that error is raised on unknown files."""
        from satpy.readers.core.grouping import group_files
        with pytest.raises(ValueError, match="No matching readers found for these files: .*"):
            group_files(self.unknown_files, "abi_l1b")

    def test_bad_reader(self):
        """Test that reader not existing causes an error."""
        import yaml

        from satpy.readers.core.grouping import group_files

        # touch the file so it exists on disk
        with mock.patch("yaml.load") as load:
            load.side_effect = yaml.YAMLError("Import problems")
            with pytest.raises(yaml.YAMLError):
                group_files([], reader="abi_l1b")

    def test_default_behavior(self):
        """Test the default behavior with the 'abi_l1b' reader."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.g16_files, reader="abi_l1b")
        assert 6 == len(groups)
        assert 2 == len(groups[0]["abi_l1b"])

    def test_default_behavior_set(self):
        """Test the default behavior with the 'abi_l1b' reader."""
        from satpy.readers.core.grouping import group_files
        files = set(self.g16_files)
        num_files = len(files)
        groups = group_files(files, reader="abi_l1b")
        # we didn't modify it
        assert len(files) == num_files
        assert 6 == len(groups)
        assert 2 == len(groups[0]["abi_l1b"])

    def test_non_datetime_group_key(self):
        """Test what happens when the start_time isn't used for grouping."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.g16_files, reader="abi_l1b", group_keys=("platform_shortname",))
        assert 1 == len(groups)
        assert 12 == len(groups[0]["abi_l1b"])

    def test_large_time_threshold(self):
        """Test what happens when the time threshold holds multiple files."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.g16_files, reader="abi_l1b", time_threshold=60*8)
        assert 3 == len(groups)
        assert 4 == len(groups[0]["abi_l1b"])

    def test_two_instruments_files(self):
        """Test the behavior when two instruments files are provided.

        This is undesired from a user point of view since we don't want G16
        and G17 files in the same Scene. Readers (like abi_l1b) are or can be
        configured to have specific group keys for handling these situations.
        Due to that this test forces the fallback group keys of
        ('start_time',).

        """
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.g16_files + self.g17_files, reader="abi_l1b", group_keys=("start_time",))
        assert 6 == len(groups)
        assert 4 == len(groups[0]["abi_l1b"])

    def test_two_instruments_files_split(self):
        """Test the default behavior when two instruments files are provided and split.

        Tell the sorting to include the platform identifier as another field
        to use for grouping.

        """
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.g16_files + self.g17_files, reader="abi_l1b",
                             group_keys=("start_time", "platform_shortname"))
        assert 12 == len(groups)
        assert 2 == len(groups[0]["abi_l1b"])
        # default for abi_l1b should also behave like this
        groups = group_files(self.g16_files + self.g17_files, reader="abi_l1b")
        assert 12 == len(groups)
        assert 2 == len(groups[0]["abi_l1b"])

    def test_viirs_orbits(self):
        """Test a reader that doesn't use 'start_time' for default grouping."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.noaa20_files + self.npp_files, reader="viirs_sdr")
        assert 2 == len(groups)
        # the noaa-20 files will be first because the orbit number is smaller
        # 5 granules * 3 file types
        assert 5 * 3 == len(groups[0]["viirs_sdr"])
        # 3 granules * 2 file types
        assert 6 == len(groups[1]["viirs_sdr"])

    def test_viirs_override_keys(self):
        """Test overriding a group keys to add 'start_time'."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(self.noaa20_files + self.npp_files, reader="viirs_sdr",
                             group_keys=("start_time", "orbit", "platform_shortname"))
        assert 8 == len(groups)
        assert 2 == len(groups[0]["viirs_sdr"])  # NPP
        assert 2 == len(groups[1]["viirs_sdr"])  # NPP
        assert 2 == len(groups[2]["viirs_sdr"])  # NPP
        assert 3 == len(groups[3]["viirs_sdr"])  # N20
        assert 3 == len(groups[4]["viirs_sdr"])  # N20
        assert 3 == len(groups[5]["viirs_sdr"])  # N20
        assert 3 == len(groups[6]["viirs_sdr"])  # N20
        assert 3 == len(groups[7]["viirs_sdr"])  # N20

        # Ask for a larger time span with our groups
        groups = group_files(self.noaa20_files + self.npp_files, reader="viirs_sdr",
                             time_threshold=60 * 60 * 2,
                             group_keys=("start_time", "orbit", "platform_shortname"))
        assert 2 == len(groups)
        # NPP is first because it has an earlier time
        # 3 granules * 2 file types
        assert 6 == len(groups[0]["viirs_sdr"])
        # 5 granules * 3 file types
        assert 5 * 3 == len(groups[1]["viirs_sdr"])

    def test_multi_readers(self):
        """Test passing multiple readers."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(
            self.g16_files + self.noaa20_files,
            reader=("abi_l1b", "viirs_sdr"))
        assert len(groups) == 11
        # test that they're grouped together when time threshold is huge and
        # only time is used to group
        groups = group_files(
            self.g16_files + self.noaa20_files,
            reader=("abi_l1b", "viirs_sdr"),
            group_keys=("start_time",),
            time_threshold=10**9)
        assert len(groups) == 1
        # test that a warning is raised when a string is passed (meaning no
        # group keys found in common)
        with pytest.warns(UserWarning, match=".*none of group keys found.*"):
            group_files(
                self.g16_files + self.noaa20_files,
                reader=("abi_l1b", "viirs_sdr"),
                group_keys=("start_time"),
                time_threshold=10**9)

    _filenames_abi_glm = [
        "OR_ABI-L1b-RadF-M6C14_G16_s19000010000000_e19000010005000_c20403662359590.nc",
        "OR_ABI-L1b-RadF-M6C14_G16_s19000010010000_e19000010015000_c20403662359590.nc",
        "OR_ABI-L1b-RadF-M6C14_G16_s19000010020000_e19000010025000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010000000_e19000010001000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010001000_e19000010002000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010002000_e19000010003000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010003000_e19000010004000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010004000_e19000010005000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010005000_e19000010006000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010006000_e19000010007000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010007000_e19000010008000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010008000_e19000010009000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010009000_e19000010010000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010010000_e19000010011000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010011000_e19000010012000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010012000_e19000010013000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010013000_e19000010014000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010014000_e19000010015000_c20403662359590.nc",
        "OR_GLM-L2-GLMF-M3_G16_s19000010015000_e19000010016000_c20403662359590.nc"]

    def test_multi_readers_empty_groups_raises_filenotfounderror(self):
        """Test behaviour on empty groups passing multiple readers.

        Make sure it raises an exception, for there will be groups
        containing GLM but not ABI.
        """
        from satpy.readers.core.grouping import group_files
        with pytest.raises(
                FileNotFoundError, match="when grouping files, group at index 1 "
                "had no files for readers: abi_l1b"):
            group_files(
                self._filenames_abi_glm,
                reader=["abi_l1b", "glm_l2"],
                group_keys=("start_time",),
                time_threshold=35,
                missing="raise")

    def test_multi_readers_empty_groups_missing_skip(self):
        """Verify empty groups are skipped.

        Verify that all groups lacking ABI are skipped, resulting in only
        three groups that are all non-empty for both instruments.
        """
        from satpy.readers.core.grouping import group_files
        groups = group_files(
            self._filenames_abi_glm,
            reader=["abi_l1b", "glm_l2"],
            group_keys=("start_time",),
            time_threshold=35,
            missing="skip")
        assert len(groups) == 2
        for g in groups:
            assert g["abi_l1b"]
            assert g["glm_l2"]

    def test_multi_readers_empty_groups_passed(self):
        """Verify that all groups are there, resulting in some that are empty."""
        from satpy.readers.core.grouping import group_files
        groups = group_files(
            self._filenames_abi_glm,
            reader=["abi_l1b", "glm_l2"],
            group_keys=("start_time",),
            time_threshold=35,
            missing="pass")
        assert len(groups) == 17
        assert not groups[1]["abi_l1b"]  # should be empty
        assert groups[1]["glm_l2"]  # should not be empty

    def test_multi_readers_invalid_parameter(self):
        """Verify that invalid missing parameter raises ValueError."""
        from satpy.readers.core.grouping import group_files
        with pytest.raises(ValueError, match="Invalid value for ``missing`` argument..*"):
            group_files(
                self._filenames_abi_glm,
                reader=["abi_l1b", "glm_l2"],
                group_keys=("start_time",),
                time_threshold=35,
                missing="hopkin green frog")


def _generate_random_string():
    import uuid
    return str(uuid.uuid1())


def _assert_is_open_file_and_close(opened):
    try:
        assert hasattr(opened, "tell")
    finally:
        opened.close()


def _posixify_path(filename):
    drive, driveless_name = os.path.splitdrive(filename)
    return driveless_name.replace("\\", "/")


@pytest.fixture(scope="module")
def random_string():
    """Random string to be used as fake file content."""
    return _generate_random_string()


@pytest.fixture(scope="module")
def local_filename(tmp_path_factory, random_string):
    """Create simple on-disk file."""
    with _local_file(tmp_path_factory, random_string) as local_path:
        yield local_path


@contextlib.contextmanager
def _local_file(tmp_path_factory, filename: str) -> Iterator[Path]:
    tmp_path = tmp_path_factory.mktemp("local_files")
    local_filename = tmp_path / filename
    local_filename.touch()
    yield local_filename


@pytest.fixture(scope="module")
def local_file(local_filename):
    """Open local file with fsspec."""
    import fsspec

    return fsspec.open(local_filename)


@pytest.fixture(scope="module")
def local_filename2(tmp_path_factory):
    """Create a second local file."""
    random_string2 = _generate_random_string()
    with _local_file(tmp_path_factory, random_string2) as local_path:
        yield local_path


@pytest.fixture(scope="module")
def local_zip_file(local_filename2):
    """Create local zip file containing one local file."""
    import zipfile

    zip_name = Path(str(local_filename2) + ".zip")
    zip_file = zipfile.ZipFile(zip_name, "w", zipfile.ZIP_DEFLATED)
    zip_file.write(local_filename2)
    zip_file.close()
    return zip_name


class TestFSFile:
    """Test the FSFile class."""

    def test_regular_filename_is_returned_with_str(self, random_string):
        """Test that str give the filename."""
        from satpy.readers.core.remote import FSFile
        assert str(FSFile(random_string)) == random_string

    def test_fsfile_with_regular_filename_abides_pathlike(self, random_string):
        """Test that FSFile abides PathLike for regular filenames."""
        from satpy.readers.core.remote import FSFile
        assert os.fspath(FSFile(random_string)) == random_string

    def test_fsfile_with_regular_filename_and_fs_spec_abides_pathlike(self, random_string):
        """Test that FSFile abides PathLike for filename+fs instances."""
        from satpy.readers.core.remote import FSFile
        assert os.fspath(FSFile(random_string, fs=None)) == random_string

    def test_fsfile_with_pathlike(self, local_filename):
        """Test FSFile with path-like object."""
        from pathlib import Path

        from satpy.readers.core.remote import FSFile
        f = FSFile(Path(local_filename))
        assert str(f) == os.fspath(f) == str(local_filename)

    def test_fsfile_with_fs_open_file_abides_pathlike(self, local_file, random_string):
        """Test that FSFile abides PathLike for fsspec OpenFile instances."""
        from satpy.readers.core.remote import FSFile
        assert os.fspath(FSFile(local_file)).endswith(random_string)

    def test_repr_includes_filename(self, local_file, random_string):
        """Test that repr includes the filename."""
        from satpy.readers.core.remote import FSFile
        assert random_string in repr(FSFile(local_file))

    def test_open_regular_file(self, local_filename):
        """Test opening a regular file."""
        from satpy.readers.core.remote import FSFile
        _assert_is_open_file_and_close(FSFile(local_filename).open())

    def test_open_local_fs_file(self, local_file):
        """Test opening a localfs file."""
        from satpy.readers.core.remote import FSFile
        _assert_is_open_file_and_close(FSFile(local_file).open())

    def test_open_zip_fs_regular_filename(self, local_filename2, local_zip_file):
        """Test opening a zipfs with a regular filename provided."""
        from fsspec.implementations.zip import ZipFileSystem

        from satpy.readers.core.remote import FSFile
        zip_fs = ZipFileSystem(local_zip_file)
        file = FSFile(_posixify_path(local_filename2), zip_fs)
        _assert_is_open_file_and_close(file.open())

    def test_open_zip_fs_openfile(self, local_filename2, local_zip_file):
        """Test opening a zipfs openfile."""
        import fsspec

        from satpy.readers.core.remote import FSFile
        open_file = fsspec.open("zip:/" + _posixify_path(local_filename2) + "::file://" + str(local_zip_file))
        file = FSFile(open_file)
        _assert_is_open_file_and_close(file.open())

    def test_sorting_fsfiles(self, local_filename, local_filename2, local_zip_file):
        """Test sorting FSFiles."""
        from fsspec.implementations.zip import ZipFileSystem

        from satpy.readers.core.remote import FSFile
        zip_fs = ZipFileSystem(local_zip_file)
        file1 = FSFile(local_filename2, zip_fs)

        file2 = FSFile(local_filename)

        extra_file = os.path.normpath("/somedir/bla")
        sorted_filenames = [os.fspath(file) for file in sorted([file1, file2, extra_file])]
        expected_filenames = sorted([extra_file, os.fspath(file1), os.fspath(file2)])
        assert sorted_filenames == expected_filenames

    def test_equality(self, local_filename, local_filename2, local_zip_file):
        """Test that FSFile compares equal when it should."""
        from fsspec.implementations.zip import ZipFileSystem

        from satpy.readers.core.remote import FSFile
        zip_fs = ZipFileSystem(local_zip_file)
        assert FSFile(local_filename) == FSFile(local_filename)
        assert (FSFile(local_filename, zip_fs) == FSFile(local_filename, zip_fs))
        assert (FSFile(local_filename, zip_fs) != FSFile(local_filename))
        assert FSFile(local_filename) != FSFile(local_filename2)

    def test_hash(self, local_filename, local_filename2, local_zip_file):
        """Test that FSFile hashing behaves sanely."""
        from fsspec.implementations.cached import CachingFileSystem
        from fsspec.implementations.local import LocalFileSystem
        from fsspec.implementations.zip import ZipFileSystem

        from satpy.readers.core.remote import FSFile

        lfs = LocalFileSystem()
        zfs = ZipFileSystem(local_zip_file)
        cfs = CachingFileSystem(fs=lfs)
        # make sure each name/fs-combi has its own hash
        assert len({hash(FSFile(fn, fs))
                    for fn in {local_filename, local_filename2}
                    for fs in [None, lfs, zfs, cfs]}) == 2 * 4

    def test_fs_property_read(self, local_filename):
        """Test reading the fs property of the class."""
        from satpy.readers.core.remote import FSFile

        fsf = FSFile(local_filename)
        fs = fsf.fs
        assert fs is None

    def test_fs_property_is_read_only(self, local_filename):
        """Test that the fs property of the class is read-only."""
        from satpy.readers.core.remote import FSFile

        fsf = FSFile(local_filename)
        with pytest.raises(AttributeError):
            fsf.fs = "foo"


def test_open_file_or_filename_uses_mode(tmp_path):
    """Test that open_file_or_filename uses provided mode."""
    from satpy.readers.core.remote import FSFile

    filename = tmp_path / "hej"
    with open(filename, mode="wb") as fd:
        fd.write(b"hej")
    fileobj = FSFile(os.fspath(filename))
    res = open_file_or_filename(fileobj, mode="rb").read()
    assert isinstance(res, bytes)


@pytest.fixture(scope="module")
def local_netcdf_filename(tmp_path_factory):
    """Create a simple local NetCDF file."""
    filename = tmp_path_factory.mktemp("fake_netcdfs") / "test.nc"
    ds = xr.Dataset()
    ds.attrs = {
        "attr1": "a",
        "attr2": 2,
    }
    ds["var1"] = xr.DataArray(np.zeros((10, 10), dtype=np.int16), dims=("y", "x"))
    ds.to_netcdf(filename)

    return str(filename)


@pytest.fixture(scope="module")
def local_netcdf_path(local_netcdf_filename):
    """Get Path object pointing to local netcdf file."""
    return Path(local_netcdf_filename)


@pytest.fixture(scope="module")
def local_netcdf_fsspec(local_netcdf_filename):
    """Get fsspec OpenFile object pointing to local netcdf file."""
    import fsspec

    return fsspec.open(local_netcdf_filename)


@pytest.fixture(scope="module")
def local_netcdf_fsfile(local_netcdf_fsspec):
    """Get FSFile object wrapping an fsspec OpenFile pointing to local netcdf file."""
    from satpy.readers.core.remote import FSFile

    return FSFile(local_netcdf_fsspec)


def _open_xarray_netcdf4():
    from functools import partial

    pytest.importorskip("netCDF4")
    return partial(xr.open_dataset, engine="netcdf4")


def _open_xarray_h5netcdf():
    from functools import partial

    pytest.importorskip("h5netcdf")
    return partial(xr.open_dataset, engine="h5netcdf")


def _open_xarray_default():
    pytest.importorskip("netCDF4")
    pytest.importorskip("h5netcdf")
    return xr.open_dataset


@pytest.fixture(scope="module")
def local_hdf5_filename(tmp_path_factory):
    """Create on-disk HDF5 file."""
    import h5py

    filename = tmp_path_factory.mktemp("fake_hdf5s") / "test.h5"
    h = h5py.File(filename, "w")
    h.create_dataset("var1", data=np.zeros((10, 10), dtype=np.int16))
    h.close()

    return str(filename)


@pytest.fixture(scope="module")
def local_hdf5_path(local_hdf5_filename):
    """Get Path object pointing to local HDF5 file."""
    return Path(local_hdf5_filename)


@pytest.fixture(scope="module")
def local_hdf5_fsspec(local_hdf5_filename):
    """Get fsspec OpenFile pointing to local HDF5 file."""
    import fsspec

    return fsspec.open(local_hdf5_filename)


def _open_h5py():
    h5py = pytest.importorskip("h5py")
    return h5py.File


@pytest.mark.parametrize(
    ("file_thing", "create_read_func"),
    [
        (lazy_fixture("local_netcdf_filename"), _open_xarray_default),
        (lazy_fixture("local_netcdf_filename"), _open_xarray_netcdf4),
        (lazy_fixture("local_netcdf_filename"), _open_xarray_h5netcdf),
        (lazy_fixture("local_netcdf_path"), _open_xarray_default),
        (lazy_fixture("local_netcdf_path"), _open_xarray_netcdf4),
        (lazy_fixture("local_netcdf_path"), _open_xarray_h5netcdf),
        (lazy_fixture("local_netcdf_fsspec"), _open_xarray_default),
        (lazy_fixture("local_netcdf_fsspec"), _open_xarray_h5netcdf),
        (lazy_fixture("local_netcdf_fsfile"), _open_xarray_default),
        (lazy_fixture("local_netcdf_fsfile"), _open_xarray_h5netcdf),
        (lazy_fixture("local_hdf5_filename"), _open_h5py),
        (lazy_fixture("local_hdf5_path"), _open_h5py),
        (lazy_fixture("local_hdf5_fsspec"), _open_h5py),
    ],
)
def test_open_file_or_filename(file_thing, create_read_func):
    """Test various combinations of file-like things and opening them with various libraries."""
    from satpy.readers.core.remote import open_file_or_filename

    read_func = create_read_func()
    open_thing = open_file_or_filename(file_thing)
    read_func(open_thing)


@pytest.mark.parametrize("name",
                         ["FSFile",
                          "open_file_or_filename",
                          "group_files",
                          "find_files_and_readers",
                          "read_reader_config",
                          "available_readers",
                          "get_valid_reader_names",
                          "OLD_READER_NAMES",
                          "PENDING_OLD_READER_NAMES",
                          "load_readers",
                          "load_reader",
                          ]
                         )
def test_init_import_warns(name):
    """Test that importing non-reader functions from __init__.py issue a warning."""
    from satpy import readers

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        _ = getattr(readers, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test fixed bugs."""


from unittest.mock import patch

import dask.array as da
import numpy as np
from xarray import DataArray, Dataset

from satpy.tests.utils import make_dataid

abi_file_list = ["/data/OR_ABI-L1b-RadF-M3C01_G16_s20180722030423_e20180722041189_c20180722041235-118900_0.nc",
                 "/data/OR_ABI-L1b-RadF-M3C02_G16_s20180722030423_e20180722041190_c20180722041228-120000_0.nc",
                 "/data/OR_ABI-L1b-RadF-M3C03_G16_s20180722030423_e20180722041190_c20180722041237-119000_0.nc",
                 "/data/OR_ABI-L1b-RadF-M3C04_G16_s20180722030423_e20180722041189_c20180722041221.nc",
                 "/data/OR_ABI-L1b-RadF-M3C05_G16_s20180722030423_e20180722041190_c20180722041237-119101_0.nc",
                 "/data/OR_ABI-L1b-RadF-M3C06_G16_s20180722030423_e20180722041195_c20180722041227.nc",
                 "/data/OR_ABI-L1b-RadF-M3C07_G16_s20180722030423_e20180722041201_c20180722041238.nc",
                 "/data/OR_ABI-L1b-RadF-M3C08_G16_s20180722030423_e20180722041190_c20180722041238.nc",
                 "/data/OR_ABI-L1b-RadF-M3C09_G16_s20180722030423_e20180722041195_c20180722041256.nc",
                 "/data/OR_ABI-L1b-RadF-M3C10_G16_s20180722030423_e20180722041201_c20180722041250.nc",
                 "/data/OR_ABI-L1b-RadF-M3C11_G16_s20180722030423_e20180722041189_c20180722041254.nc",
                 "/data/OR_ABI-L1b-RadF-M3C12_G16_s20180722030423_e20180722041195_c20180722041256.nc",
                 "/data/OR_ABI-L1b-RadF-M3C13_G16_s20180722030423_e20180722041201_c20180722041259.nc",
                 "/data/OR_ABI-L1b-RadF-M3C14_G16_s20180722030423_e20180722041190_c20180722041258.nc",
                 "/data/OR_ABI-L1b-RadF-M3C15_G16_s20180722030423_e20180722041195_c20180722041259.nc",
                 "/data/OR_ABI-L1b-RadF-M3C16_G16_s20180722030423_e20180722041202_c20180722041259.nc"]


def generate_fake_abi_xr_dataset(filename, chunks=None, **kwargs):
    """Create a fake xarray dataset for abi data.

    This is an incomplete copy of existing file structures.
    """
    dataset = Dataset(attrs={
        "time_coverage_start": "2018-03-13T20:30:42.3Z",
        "time_coverage_end": "2018-03-13T20:41:18.9Z",
    })

    projection = DataArray(
        [-214748364],
        attrs={
            "long_name": "GOES-R ABI fixed grid projection",
            "grid_mapping_name": "geostationary",
            "perspective_point_height": 35786023.0,
            "semi_major_axis": 6378137.0,
            "semi_minor_axis": 6356752.31414,
            "inverse_flattening": 298.2572221,
            "latitude_of_projection_origin": 0.0,
            "longitude_of_projection_origin": -75.0,
            "sweep_angle_axis": "x"
        })
    dataset["goes_imager_projection"] = projection

    if "C01" in filename or "C03" in filename or "C05" in filename:
        stop = 10847
        step = 2
        scale = 2.8e-05
        offset = 0.151858
    elif "C02" in filename:
        stop = 21693
        step = 4
        scale = 1.4e-05
        offset = 0.151865
    else:
        stop = 5424
        step = 1
        scale = 5.6e-05
        offset = 0.151844

    y = DataArray(
        da.arange(0, stop, step),
        attrs={
            "scale_factor": -scale,
            "add_offset": offset,
            "units": "rad",
            "axis": "Y",
            "long_name": "GOES fixed grid projection y-coordinate",
            "standard_name": "projection_y_coordinate"
            },
        dims=["y"])

    dataset["y"] = y

    x = DataArray(
        da.arange(0, stop, step),
        attrs={
            "scale_factor": scale,
            "add_offset": -offset,
            "units": "rad",
            "axis": "X",
            "long_name": "GOES fixed grid projection x-coordinate",
            "standard_name": "projection_x_coordinate"
        },
        dims=["x"])

    dataset["x"] = x

    rad = DataArray(
        da.random.randint(0, 1025, size=[len(y), len(x)], dtype=np.int16, chunks=chunks),
        attrs={
            "_FillValue": np.array(1023),
            "long_name": "ABI L1b Radiances",
            "standard_name": "toa_outgoing_radiance_per_unit_wavelength",
            "_Unsigned": "true",
            "sensor_band_bit_depth": 10,
            "valid_range": np.array([0, 1022], dtype=np.int16),
            "scale_factor": 0.8121064,
            "add_offset": -25.936647,
            "units": "W m-2 sr-1 um-1",
            "resolution": "y: 0.000028 rad x: 0.000028 rad",
            "grid_mapping": "goes_imager_projection",
            "cell_methods": "t: point area: point"
        },
        dims=["y", "x"]
    )

    dataset["Rad"] = rad

    sublat = DataArray(0.0, attrs={
        "long_name": "nominal satellite subpoint latitude (platform latitude)",
        "standard_name": "latitude",
        "_FillValue": -999.0,
        "units": "degrees_north"})
    dataset["nominal_satellite_subpoint_lat"] = sublat

    sublon = DataArray(-75.0, attrs={
        "long_name": "nominal satellite subpoint longitude (platform longitude)",
        "standard_name": "longitude",
        "_FillValue": -999.0,
        "units": "degrees_east"})

    dataset["nominal_satellite_subpoint_lon"] = sublon

    satheight = DataArray(35786.023, attrs={
        "long_name": "nominal satellite height above GRS 80 ellipsoid (platform altitude)",
        "standard_name": "height_above_reference_ellipsoid",
        "_FillValue": -999.0,
        "units": "km"})

    dataset["nominal_satellite_height"] = satheight

    yaw_flip_flag = DataArray(0, attrs={
        "long_name": "Flag indicating the spacecraft is operating in yaw flip configuration",
        "_Unsigned": "true",
        "_FillValue": np.array(-1),
        "valid_range": np.array([0, 1], dtype=np.int8),
        "units": "1",
        "flag_values": "0 1",
        "flag_meanings": "false true"})

    dataset["yaw_flip_flag"] = yaw_flip_flag

    return dataset


@patch("xarray.open_dataset")
def test_1258(fake_open_dataset):
    """Save true_color from abi with radiance doesn't need two resamplings."""
    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader="abi_l1b")
    scene.load(["true_color_nocorr", "C04"], calibration="radiance")
    resampled_scene = scene.resample(scene.coarsest_area(), resampler="native")
    assert len(resampled_scene.keys()) == 2


@patch("xarray.open_dataset")
def test_1088(fake_open_dataset):
    """Check that copied arrays gets resampled."""
    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader="abi_l1b")
    scene.load(["C04"], calibration="radiance")

    my_id = make_dataid(name="my_name", wavelength=(10, 11, 12))
    scene[my_id] = scene["C04"].copy()
    resampled = scene.resample("eurol")
    assert resampled[my_id].shape == (2048, 2560)


@patch("xarray.open_dataset")
def test_no_enums(fake_open_dataset):
    """Check that no enums are inserted in the resulting attrs."""
    from enum import Enum

    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader="abi_l1b")
    scene.load(["C04"], calibration="radiance")
    for value in scene["C04"].attrs.values():
        assert not isinstance(value, Enum)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""GCOM-W1 AMSR2 Level 2 files from the GAASP software.

GAASP output files are in the NetCDF4 format. Software is provided by NOAA
and is also distributed by the CSPP group. More information on the products
supported by this reader can be found here:
https://www.star.nesdis.noaa.gov/jpss/gcom.php for more information.

GAASP includes both swath/granule products and gridded products. Swath
products are provided in files with "MBT", "OCEAN", "SNOW", or "SOIL" in the
filename. Gridded products are in files with "SEAICE-SH" or "SEAICE-NH" in the
filename where SH stands for South Hemisphere and NH stands for North
Hemisphere. These gridded products are on the EASE2 North pole and South pole
grids. See https://nsidc.org/ease/ease-grid-projection-gt for more details.

Note that since SEAICE products can be on both the northern or
southern hemisphere or both depending on what files are provided to Satpy, this
reader appends a `_NH` and `_SH` suffix to all variable names that are
dynamically discovered from the provided files.

Data Quality Filtering
^^^^^^^^^^^^^^^^^^^^^^

Some variables can be filtered based on Quality Control (QC) variables in the
data files. At the time of writing the only supported filtering is the "WSPD"
variable from the "OCEAN" files using the "WSPD_QC" variable. Where the QC
variable is non-zero the WSPD data is set to NaN. By default no filtering is
applied. To enable filtering pass the keyword arguments to the ``Scene``::

    scn = Scene(reader="amsr2_l2_gaasp", filenames=[...],
                reader_kwargs={"filter_wind_speed": True})

Note that many variables in the GAASP data files already have additional
quality filtering applied by the algorithm.

"""

import datetime as dt
import logging
from typing import Tuple

import numpy as np
import xarray as xr
from pyproj import CRS
from pyresample.geometry import AreaDefinition

from satpy._compat import cached_property
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()


class GAASPFileHandler(BaseFileHandler):
    """Generic file handler for GAASP output files."""

    y_dims: Tuple[str, ...] = (
        "Number_of_Scans",
    )
    x_dims: Tuple[str, ...] = (
        "Number_of_hi_rez_FOVs",
        "Number_of_low_rez_FOVs",
    )
    time_dims = (
        "Time_Dimension",
    )
    is_gridded = False
    dim_resolutions = {
        "Number_of_hi_rez_FOVs": 5000,
        "Number_of_low_rez_FOVs": 10000,
    }

    def __init__(self, filename, filename_info, filetype_info, filter_wind_speed=False):
        """Initialize file reading and store filter keyword arguments."""
        super().__init__(filename, filename_info, filetype_info)
        self.filter_wind_speed = filter_wind_speed

    @cached_property
    def nc(self):
        """Get the xarray dataset for this file."""
        chunks = {dim_name: CHUNK_SIZE for dim_name in
                  self.y_dims + self.x_dims + self.time_dims}
        nc = xr.open_dataset(self.filename,
                             decode_cf=True,
                             decode_timedelta=False,
                             mask_and_scale=False,
                             chunks=chunks)

        if len(self.time_dims) == 1:
            nc = nc.rename({self.time_dims[0]: "time"})
        return nc

    @property
    def start_time(self):
        """Get start time of observation."""
        try:
            return self.filename_info["start_time"]
        except KeyError:
            time_str = self.nc.attrs["time_coverage_start"]
            return dt.datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get end time of observation."""
        try:
            return self.filename_info["end_time"]
        except KeyError:
            time_str = self.nc.attrs["time_coverage_end"]
            return dt.datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def sensor_names(self):
        """Sensors who have data in this file."""
        return {self.nc.attrs["instrument_name"].lower()}

    @property
    def platform_name(self):
        """Name of the platform whose data is stored in this file."""
        return self.nc.attrs["platform_name"]

    def _get_var_name_without_suffix(self, var_name):
        var_suffix = self.filetype_info.get("var_suffix", "")
        if var_suffix:
            var_name = var_name[:-len(var_suffix)]
        return var_name

    def _scale_data(self, data_arr, attrs):
        # handle scaling
        # take special care for integer/category fields
        scale_factor = attrs.pop("scale_factor", 1.)
        add_offset = attrs.pop("add_offset", 0.)
        scaling_needed = not (scale_factor == 1 and add_offset == 0)
        if scaling_needed:
            new_dtype = np.float32 if np.issubdtype(data_arr.dtype, np.integer) else data_arr.dtype.type
            data_arr = data_arr * new_dtype(scale_factor) + new_dtype(add_offset)
        return data_arr, attrs

    @staticmethod
    def _nan_for_dtype(data_arr_dtype):
        # don't force the conversion from 32-bit float to 64-bit float
        # if we don't have to
        if data_arr_dtype.type == np.float32:
            return np.float32(np.nan)
        if np.issubdtype(data_arr_dtype, np.timedelta64):
            return np.timedelta64("NaT")
        if np.issubdtype(data_arr_dtype, np.datetime64):
            return np.datetime64("NaT")
        return np.nan

    def _fill_data(self, data_arr, attrs):
        fill_value = attrs.pop("_FillValue", None)
        is_int = np.issubdtype(data_arr.dtype, np.integer)
        has_flag_comment = "comment" in attrs
        if is_int and has_flag_comment:
            # category or timedelta product
            fill_out = fill_value
            attrs["_FillValue"] = fill_out
        else:
            fill_out = self._nan_for_dtype(data_arr.dtype)
        if fill_value is not None:
            data_arr.data.compute().astype("float32")
            data_arr = data_arr.where(data_arr != fill_value, fill_out)
        return data_arr, attrs

    def get_dataset(self, dataid, ds_info):
        """Load, scale, and collect metadata for the specified DataID."""
        orig_var_name = self._get_var_name_without_suffix(dataid["name"])
        data_arr = self.nc[orig_var_name].copy()
        attrs = data_arr.attrs.copy()
        data_arr, attrs = self._scale_data(data_arr, attrs)
        data_arr, attrs = self._fill_data(data_arr, attrs)
        data_arr = self._filter_by_qc(orig_var_name, data_arr)

        attrs.update({
            "platform_name": self.platform_name,
            "sensor": sorted(self.sensor_names)[0],
            "start_time": self.start_time,
            "end_time": self.end_time,
        })
        dim_map = dict(zip(data_arr.dims, ("y", "x")))
        # rename dims
        data_arr = data_arr.rename(**dim_map)
        # drop coords, the base reader will recreate these
        data_arr = data_arr.reset_coords(drop=True)
        data_arr.attrs = attrs
        return data_arr

    def _filter_by_qc(self, orig_var_name: str, data_arr: xr.DataArray) -> xr.DataArray:
        if orig_var_name == "WSPD" and self.filter_wind_speed:
            wspd_qc = self.nc["WSPD_QC"]
            data_arr = data_arr.where(wspd_qc == 0)
        return data_arr


    def _available_if_this_file_type(self, configured_datasets):
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            yield self.file_type_matches(ds_info["file_type"]), ds_info

    def _add_lonlat_coords(self, data_arr, ds_info):
        lat_coord = None
        lon_coord = None
        for coord_name in data_arr.coords:
            if "longitude" in coord_name.lower():
                lon_coord = coord_name
            if "latitude" in coord_name.lower():
                lat_coord = coord_name
        ds_info["coordinates"] = [lon_coord, lat_coord]

    def _get_ds_info_for_data_arr(self, var_name, data_arr):
        var_suffix = self.filetype_info.get("var_suffix", "")
        ds_info = {
            "file_type": self.filetype_info["file_type"],
            "name": var_name + var_suffix,
        }
        x_dim_name = data_arr.dims[1]
        if x_dim_name in self.dim_resolutions:
            ds_info["resolution"] = self.dim_resolutions[x_dim_name]
        if not self.is_gridded and data_arr.coords:
            self._add_lonlat_coords(data_arr, ds_info)
        return ds_info

    def _is_2d_yx_data_array(self, data_arr):
        has_y_dim = data_arr.dims[0] in self.y_dims
        has_x_dim = data_arr.dims[1] in self.x_dims
        return has_y_dim and has_x_dim

    def _available_new_datasets(self):
        possible_vars = list(self.nc.data_vars.items()) + list(self.nc.coords.items())
        for var_name, data_arr in possible_vars:
            if data_arr.ndim != 2:
                # we don't currently handle non-2D variables
                continue
            if not self._is_2d_yx_data_array(data_arr):
                # we need 'traditional' y/x dimensions currently
                continue

            ds_info = self._get_ds_info_for_data_arr(var_name, data_arr)
            yield True, ds_info

    def available_datasets(self, configured_datasets=None):
        """Dynamically discover what variables can be loaded from this file.

        See :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for more information.

        """
        yield from self._available_if_this_file_type(configured_datasets)
        yield from self._available_new_datasets()


class GAASPGriddedFileHandler(GAASPFileHandler):
    """GAASP file handler for gridded products like SEAICE."""

    y_dims = (
        "Number_of_Y_Dimension",
    )
    x_dims = (
        "Number_of_X_Dimension",
    )
    dim_resolutions = {
        "Number_of_X_Dimension": 10000,
    }
    is_gridded = True

    @staticmethod
    def _get_extents(data_shape, res):
        # assume data is centered at projection center
        x_min = -(data_shape[1] / 2.0) * res
        x_max = (data_shape[1] / 2.0) * res
        y_min = -(data_shape[0] / 2.0) * res
        y_max = (data_shape[0] / 2.0) * res
        return x_min, y_min, x_max, y_max

    def get_area_def(self, dataid):
        """Create area definition for equirectangular projected data."""
        var_suffix = self.filetype_info.get("var_suffix", "")
        area_name = "gaasp{}".format(var_suffix)
        orig_var_name = self._get_var_name_without_suffix(dataid["name"])
        data_shape = self.nc[orig_var_name].shape
        crs = CRS(self.filetype_info["grid_epsg"])
        res = dataid["resolution"]
        extent = self._get_extents(data_shape, res)
        area_def = AreaDefinition(
            area_name,
            area_name,
            area_name,
            crs,
            data_shape[1],
            data_shape[0],
            extent
        )
        return area_def


class GAASPLowResFileHandler(GAASPFileHandler):
    """GAASP file handler for files that only have low resolution products."""

    x_dims = (
        "Number_of_low_rez_FOVs",
    )
    dim_resolutions = {
        "Number_of_low_rez_FOVs": 10000,
    }
